#!/usr/bin/env python
"""@package GenerateSoilTextureMap

@brief Query RHESSys ParamDB for soil definitions for a custom soil map.  

This software is provided free of charge under the New BSD License. Please see
the following license information:

Copyright (c) 2013, University of North Carolina at Chapel Hill
All rights reserved.

Redistribution and use in source and binary forms, with or without
modification, are permitted provided that the following conditions are met:
    * Redistributions of source code must retain the above copyright
      notice, this list of conditions and the following disclaimer.
    * Redistributions in binary form must reproduce the above copyright
      notice, this list of conditions and the following disclaimer in the
      documentation and/or other materials provided with the distribution.
    * Neither the name of the University of North Carolina at Chapel Hill nor the
      names of its contributors may be used to endorse or promote products
      derived from this software without specific prior written permission.

THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND
ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
DISCLAIMED. IN NO EVENT SHALL THE UNIVERSITY OF NORTH CAROLINA AT CHAPEL HILL
BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR 
CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE
GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT 
LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT
OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.


@author Brian Miles <brian_miles@unc.edu>


Pre conditions
--------------
1. Configuration file must define the following sections and values:
   'GRASS', 'GISBASE'

2. The following metadata entry(ies) must be present in the GRASS section of the metadata associated with the project directory:
   dem_rast
   soil_rast

3. The following metadata entry(ies) must be present in the RHESSys section of the metadata associated with the project directory:
   paramdb
   paramdb_dir
   grass_dbase
   grass_location
   grass_mapset
   rhessys_dir

Post conditions
---------------
1. Will write the following entry(ies) to the RHESSys section of metadata associated with the project directory:
   soil_rule
   
Usage:
@code
GenerateCustomSoilDefinitions.py -p /path/to/project_dir
@endcode

@note EcoHydroWorkflowLib configuration file must be specified by environmental variable 'ECOHYDROWORKFLOW_CFG',
or -i option must be specified. 
"""
import os, sys, errno
import argparse
import textwrap

from ecohydrolib.grasslib import *

from rhessysworkflows.context import Context
from rhessysworkflows.metadata import RHESSysMetadata
from rhessysworkflows.rhessys import RHESSysPaths

# Handle command line options
parser = argparse.ArgumentParser(description='Generate soil texture map for dataset in GRASS GIS')
parser.add_argument('-i', '--configfile', dest='configfile', required=False,
                    help='The configuration file. Must define section "GRASS" and option "GISBASE"')
parser.add_argument('-p', '--projectDir', dest='projectDir', required=True,
                    help='The directory to which metadata, intermediate, and final files should be saved')
group = parser.add_mutually_exclusive_group(required=True)
group.add_argument('-b', '--buildPrototypeRules', dest='buildPrototypeRules', required=False, action='store_true',
                    help='Write prototype soil reclass rules to the project directory. You must edit these rules to match the classes in your custom soils data')
group.add_argument('-r', '--ruleDir', dest='ruleDir', required=False,
                    help="The directory where soils reclass rule can be found; must contain the file %s" % (str(RHESSysMetadata.SOILS_RULES),) )
args = parser.parse_args()
cmdline = RHESSysMetadata.getCommandLine()

configFile = None
if args.configfile:
    configFile = args.configfile

context = Context(args.projectDir, configFile) 

# Check for necessary information in metadata
metadata = RHESSysMetadata.readRHESSysEntries(context)
if not 'grass_dbase' in metadata:
    sys.exit("Metadata in project directory %s does not contain a GRASS Dbase" % (context.projectDir,)) 
if not 'grass_location' in metadata:
    sys.exit("Metadata in project directory %s does not contain a GRASS location" % (context.projectDir,)) 
if not 'grass_mapset' in metadata:
    sys.exit("Metadata in project directory %s does not contain a GRASS mapset" % (context.projectDir,))
if not 'paramdb_dir' in metadata:
    sys.exit("Metadata in project directory %s does not contain a ParamDB directory" % (context.projectDir,))
if not 'paramdb' in metadata:
    sys.exit("Metadata in project directory %s does not contain a ParamDB" % (context.projectDir,))

grassMetadata = RHESSysMetadata.readGRASSEntries(context)
if not 'dem_rast' in grassMetadata:
    sys.exit("Metadata in project directory %s does not contain a DEM raster in a GRASS mapset" % (context.projectDir,)) 
demRast = grassMetadata['dem_rast']
if not 'soil_rast' in grassMetadata:
    sys.exit("Metadata in project directory %s does not contain a soil raster in a GRASS mapset" % (context.projectDir,)) 
soilRast = grassMetadata['soil_rast']

paths = RHESSysPaths(args.projectDir, metadata['rhessys_dir'])

# Write prototype soil reclass rules to the project directory
if args.buildPrototypeRules:
    sys.stdout.write('Generating prototype soil reclass rules...')
    sys.stdout.flush()
    # Get path of place to store reclass rules
    projectDirRuleDir = paths.getReclassRulesDirectory()
    soilsRulePath = os.path.join(projectDirRuleDir, RHESSysMetadata.SOILS_RULE)
    with open(soilsRulePath, 'w') as f:
        f.write('1 = 1 clay\n')
        f.write('2 = 2 silt-clay\n')
        f.write('3 = 3 silt-clay-loam\n')
        f.write('4 = 4 sandy-clay\n')
        f.write('5 = 5 sandy-clay-loam\n')
        f.write('6 = 6 clay-loam\n')
        f.write('7 = 7 silt\n')
        f.write('8 = 8 silt-loam\n')
        f.write('9 = 9 loam\n')
        f.write('10 = 10 sand\n')
        f.write('11 = 11 loamy-sand\n')
        f.write('12 = 12 sandy-loam\n')
        f.write('13 = 13 rock\n')
        f.write('14 = 14 water\n')
    ruleDir = None
    sys.stdout.write('done\n')
    

# Import rules from elsewhere
if args.ruleDir:
    if not os.access(args.ruleDir, os.R_OK):
        sys.exit("Unable to read rule directory %s" % (ruleDir,) )
    ruleDir = os.path.abspath(args.ruleDir)

if ruleDir:
    sys.stdout.write(textwrap.fill("Importing soil reclass rules from %s ..." % (ruleDir,) ) )
    sys.stdout.flush()
    # Copy rules into project directory
    soilsRulePath = os.path.join(ruleDir, RHESSysMetadata.SOILS_RULE)
    shutil.copy(soilsRulePath, projectDirRuleDir)
    sys.stdout.write('done\n')
    
# Write metadata
RHESSysMetadata.writeRHESSysEntry(context, 'soil_rule', os.path.join(RHESSysMetadata.RULES_DIR, os.path.basename(soilsRulePath)) )

# Write processing history
RHESSysMetadata.appendProcessingHistoryItem(context, cmdline)


#!/usr/bin/env python
"""@package ImportClimateData

@brief Import climate station and corrsponding data, already in RHESSys format, 
into project directory

This software is provided free of charge under the New BSD License. Please see
the following license information:

Copyright (c) 2013, University of North Carolina at Chapel Hill
All rights reserved.

Redistribution and use in source and binary forms, with or without
modification, are permitted provided that the following conditions are met:
    * Redistributions of source code must retain the above copyright
      notice, this list of conditions and the following disclaimer.
    * Redistributions in binary form must reproduce the above copyright
      notice, this list of conditions and the following disclaimer in the
      documentation and/or other materials provided with the distribution.
    * Neither the name of the University of North Carolina at Chapel Hill nor the
      names of its contributors may be used to endorse or promote products
      derived from this software without specific prior written permission.

THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND
ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
DISCLAIMED. IN NO EVENT SHALL THE UNIVERSITY OF NORTH CAROLINA AT CHAPEL HILL
BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR 
CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE
GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT 
LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT
OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.


@author Brian Miles <brian_miles@unc.edu>


Pre conditions
--------------
1. The following metadata entry(ies) must be present in the RHESSys section of the metadata associated with the project directory:
   rhessys_dir
   
Post conditions
---------------
1. Will write the following entry(ies) to the RHESSys section of metadata associated with the project directory:
   climate_stations


Usage:
@code
ImportClimateData.py -p /path/to/project_dir -s /path/to/existing/rhessys/climate/data
@endcode
"""
import os, sys, errno
import argparse
import re
import shutil
import textwrap

from rhessysworkflows.context import Context
from rhessysworkflows.metadata import RHESSysMetadata
from rhessysworkflows.rhessys import RHESSysPaths

STATION_FILE_EXTENSION = 'base'

# Handle command line options
parser = argparse.ArgumentParser(description='Import RHESSys climate data into project directory')
parser.add_argument('-i', '--configfile', dest='configfile', required=False,
                    help='The configuration file. Must define section "GRASS" and option "GISBASE"')
parser.add_argument('-p', '--projectDir', dest='projectDir', required=True,
                    help='The directory to which metadata, intermediate, and final files should be saved')
parser.add_argument('-s', '--sourceDir', dest='sourceDir', required=True,
                    help='The directory from which climate data should be copied. Assumes climate base stations are stored in files ending in ".base", with climate data files of the form ".VARNAME" (e.g. ".tmin")')
parser.add_argument('--overwrite', dest='overwrite', action='store_true', required=False,
                    help='Overwrite existing climate stations and datasets; If specified, will delete existing data before importing new data.  If not specified, new data will be added to existing data.')
args = parser.parse_args()
cmdline = RHESSysMetadata.getCommandLine()

configFile = None
if args.configfile:
    configFile = args.configfile

context = Context(args.projectDir, configFile)

# Check for necessary information in metadata
metadata = RHESSysMetadata.readRHESSysEntries(context)
paths = RHESSysPaths(args.projectDir, metadata['rhessys_dir'])

if not os.access(args.sourceDir, os.R_OK):
    sys.exit("The specified path of the climate data source directory, %s, is not readable" % (args.sourceDir,) )
if not os.path.isdir(args.sourceDir):
    sys.exit("The specified path of the climate data source directory, %s, is not a directory" % (args.sourceDir,) )
    
if args.overwrite:
    # Delete any existing climate data
    contents = os.listdir(paths.RHESSYS_CLIM)
    for entry in contents:
        fileToDelete = os.path.join(paths.RHESSYS_CLIM, entry)
        print("Deleting existing climate file %s" % (fileToDelete,) )
        os.unlink(fileToDelete)
    
# Get listing of climate stations in source directory
stationRegex = re.compile("^(.+)\.%s$" % (STATION_FILE_EXTENSION,) )
contents = os.listdir(args.sourceDir)
stations = []
for entry in contents:
    m = stationRegex.match(entry)
    if m:
        stations.append(m.group(1))    

# Copy station file and climate data
sys.stdout.write(textwrap.fill("Importing climate data from %s...\n" % (args.sourceDir,) ) )
sys.stdout.flush()
for station in stations:
    # Copy station file
    sys.stdout.write("\n\tStation '%s'\n" % (station,) )
    sys.stdout.flush()
    stationFileName = "%s%s%s" % (station, os.extsep, STATION_FILE_EXTENSION)
    stationFilePath = os.path.join(args.sourceDir, stationFileName)
    shutil.copy(stationFilePath, paths.RHESSYS_CLIM)
    dataRegex = re.compile("^%s\.(.+)$" % (station,) )
    # Look for climate data files to copy and copy them
    for entry in contents:
        m = dataRegex.match(entry)
        if m:
            extension = m.group(1)
            if extension != STATION_FILE_EXTENSION:
                shutil.copy( os.path.join(args.sourceDir, entry), paths.RHESSYS_CLIM )
sys.stdout.write('done\n')

# Write metadata
RHESSysMetadata.writeRHESSysEntry(context, 'climate_stations', RHESSysMetadata.VALUE_DELIM.join(stations))

# Write processing history
RHESSysMetadata.appendProcessingHistoryItem(context, cmdline)
  


#!/usr/bin/env python
"""@package RHESSysPlotMassbalance

@brief Tool for comparing mass balance for observed and basin-scale RHESSys output.

This software is provided free of charge under the New BSD License. Please see
the following license information:

Copyright (c) 2014, University of North Carolina at Chapel Hill
All rights reserved.

Redistribution and use in source and binary forms, with or without
modification, are permitted provided that the following conditions are met:
    * Redistributions of source code must retain the above copyright
      notice, this list of conditions and the following disclaimer.
    * Redistributions in binary form must reproduce the above copyright
      notice, this list of conditions and the following disclaimer in the
      documentation and/or other materials provided with the distribution.
    * Neither the name of the University of North Carolina at Chapel Hill nor the
      names of its contributors may be used to endorse or promote products
      derived from this software without specific prior written permission.

THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND
ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
DISCLAIMED. IN NO EVENT SHALL THE UNIVERSITY OF NORTH CAROLINA AT CHAPEL HILL
BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR 
CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE
GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT 
LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT
OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.


@author Brian Miles <brian_miles@unc.edu>

"""
import sys
import argparse
import math

import numpy as np
import pandas as pd
import statsmodels.api as sm
import matplotlib.pyplot as plt
import matplotlib

from rhessysworkflows.rhessys import RHESSysOutput


OBS_HEADER_STREAMFLOW = 'streamflow_mm'
OBS_HEADER_PRECIP = 'precip_mm'

PLOT_TYPE_STD = 'standard'
PLOT_TYPE_LOGY = 'logy'
PLOT_TYPE_CDF = 'cdf'
PLOT_TYPE_TABLE = 'table'
PLOT_TYPES = [PLOT_TYPE_STD, PLOT_TYPE_LOGY, PLOT_TYPE_CDF, PLOT_TYPE_TABLE]
PLOT_DEFAULT = PLOT_TYPE_STD


def plotTable(args, col_names, obs, data, ax):
    startDate = data['streamflow'].index[0]
    endDate = data['streamflow'].index[-1]
    
    obsRunoffRat = np.sum( obs[OBS_HEADER_STREAMFLOW] ) / np.sum( obs[OBS_HEADER_PRECIP] )
    modRunoffRat = np.sum( data['streamflow']) / np.sum( data['precip'] )
    obsET = np.sum( obs[OBS_HEADER_PRECIP] ) - np.sum( obs[OBS_HEADER_STREAMFLOW] )
    summary = [ [ np.sum( obs[OBS_HEADER_PRECIP] ), np.sum( data['precip'] ) ],
                [ np.sum(obs[OBS_HEADER_STREAMFLOW]), np.sum(data['streamflow']) ],
                [ obsET, np.sum( data['evap'] ) + np.sum( data['trans'] ) ],
                [ obsRunoffRat, modRunoffRat ]
              ]
    text = [ [ "%.2f" % num for num in summary[0] ],
             [ "%.2f" % num for num in summary[1] ],
             [ "%.2f" % num for num in summary[2] ],
             [ "%.2f" % num for num in summary[3] ]
           ]

    ax.axis('off')
    # Draw the table
    table = ax.table(cellText=text,
                     colWidths=[0.33, 0.33],
                     colLabels=col_names,
                     rowLabels=['Precip ($mm$)', 'Streamflow ($mm$)', 'ET ($mm$)', 'Runoff ratio'],
                     loc='center right')
    table.auto_set_font_size(False)
    table.set_fontsize(9)
    ax.text(0, 0.75, "Mass balance: %s - %s" % \
                 (startDate.strftime('%Y/%m/%d'), endDate.strftime('%Y/%m/%d') ),
            horizontalalignment='left',
            #verticalalignment='center',
            fontsize=10)
    

def plotGraph(args, plottype, obs, data, columns, min_x, max_x, ax, secondary=None,
              plotColor=True):
    
    if plotColor:
        obs_color = 'black'
        mod_color = 'green'
        second_color = 'blue'
    else:
        obs_color = 'black'
        mod_color = 'grey'
        second_color = 'black'
    
    if plottype == PLOT_TYPE_STD or \
       plottype == PLOT_TYPE_LOGY:
        x = obs.index
    elif plottype == PLOT_TYPE_CDF:
        x = np.linspace(min_x, max_x, num=1000 )
    
    # Plot observed values
    # Standard or log plot
    obs_y = obs[OBS_HEADER_STREAMFLOW]
    if plottype == PLOT_TYPE_CDF:
        obs_ecdf = sm.distributions.ECDF(obs_y)
        obs_y = obs_ecdf(x)
    obs_plt = None
    if not args.supressObs:
        (obs_plt,) = ax.plot(x, obs_y, obs_color, linewidth=2)
        
    # Plot modeled values
    data_plt = []
    for c in columns:
        # Standard or log plot
        mod_y = data[c]
        if plottype == PLOT_TYPE_CDF:
            mod_ecdf = sm.distributions.ECDF(data[c])
            mod_y = mod_ecdf(x)
        (mod_plt,) = ax.plot(x, mod_y, color=mod_color, linewidth=1)
        data_plt.append(mod_plt)
    
    # X-axis
    if plottype == PLOT_TYPE_STD or \
       plottype == PLOT_TYPE_LOGY:
        num_years = len(x) / 365
        if num_years > 2:
            if num_years > 5:
                ax.xaxis.set_major_locator(matplotlib.dates.YearLocator())
            else:
                ax.xaxis.set_major_locator(matplotlib.dates.MonthLocator(interval=3))
        else:
            ax.xaxis.set_major_locator(matplotlib.dates.MonthLocator())
        ax.xaxis.set_major_formatter(matplotlib.dates.DateFormatter('%b-%Y'))
        # Rotate
        plt.setp( ax.xaxis.get_majorticklabels(), rotation=45 )
        plt.setp( ax.xaxis.get_majorticklabels(), fontsize='x-small' )
    
    if plottype == PLOT_TYPE_CDF:
        ax.set_xlim(min_x, max_x)
        ax.set_xscale('log')
        if args.xlabel:
            ax.set_xlabel(args.xlabel)
        else:
            pass
            ax.set_xlabel( 'Streamflow ($mm\ day^{-1}$)' )
    elif args.xlabel:
        ax.set_xlabel(args.xlabel)
    
    # Y-axis
    if plottype == PLOT_TYPE_LOGY:
        ax.set_yscale('log')
    
    if args.ylabel:
        ax.set_ylabel(args.ylabel)
    elif plottype != PLOT_TYPE_CDF and plottype != PLOT_TYPE_LOGY:
        y_label = 'Streamflow ($mm\ day^{-1}$)'
        ax.set_ylabel( y_label )
    
    data_plt.insert(0, obs_plt)
    
    # Plot secondary data (if specified)
    if secondary and \
        (plottype == PLOT_TYPE_STD or plottype == PLOT_TYPE_LOGY):
        # Plot
        ax2 = ax.twinx()
        if args.secondaryPlotType == 'line':
            (sec_plot,) = ax2.plot(x, data[secondary])
        elif args.secondaryPlotType == 'bar':
            sec_plot = ax2.bar(x, data[secondary], facecolor=second_color, edgecolor='none', width=2.0)
        ax2.invert_yaxis()
        ax2.set_ylabel('Precipication ($mm\ day^{-1}$)')
    
    return data_plt

if __name__ == "__main__":
    # Handle command line options
    parser = argparse.ArgumentParser(description='Plot CDF of N datasets vs. observed data')
    parser.add_argument('-n', '--outname', required=True, 
                        help='Base name of file to output figure to.  Only specify base name of file, not extension (PDF and PNG files will be produced)')
    parser.add_argument('-o', '--obs', required=True,
                        help='File containing observed data.  Must be a CSV file with the following headers: datetime, streamflow_mm, precip_mm')
    parser.add_argument('-d', '--data', required=True,
                        help='One or more data files')
    parser.add_argument('-t', '--title', required=False,
                        help='Title of figure')
    parser.add_argument('-x', '--xlabel', required=False,
                        help='X-axis label')
    parser.add_argument('-y', '--ylabel', required=False,
                        help='Y-axis label')
    parser.add_argument('--figureX', required=False, type=int, default=8,
                        help='The width of the plot, in inches')
    parser.add_argument('--figureY', required=False, type=int, default=6,
                        help='The height of the plot, in inches')
    parser.add_argument('--supressObs', required=False, action='store_true',
                        help='Do not plot observed data.  Observed data will still be used for aligning timeseries')
    parser.add_argument("--color", action="store_true", required=False, default=False,
                        help="Plot in color")
    parser.add_argument('--secondaryPlotType', required=False, choices=['bar', 'line'], default='bar',
                        help='Type of plot to use for secondary data.')
    parser.add_argument('--secondaryLabel', required=False,
                        help='Label to use for seconary Y-axis')
    args = parser.parse_args()

    # Open observed data
    obs = pd.read_csv(args.obs, index_col=0, parse_dates=True)

    # Open data and align to observed
    cols = ['streamflow', 'evap', 'trans', 'precip']
    obs_align = None
    max_x = min_x = 0
    mod_file = open(args.data, 'r')
    mod_df = RHESSysOutput.readColumnsFromFile(mod_file, cols,
                                               readHour=False)
    
    # Align timeseries
    (mod_align, obs_align) = mod_df.align(obs, axis=0, join='inner')
    tmp_max_x = max(mod_align['streamflow'].max(), obs_align[OBS_HEADER_STREAMFLOW].max())
    if tmp_max_x > max_x:
        max_x = tmp_max_x
    min_x = max(min_x, mod_align['streamflow'].min())

    mod_file.close()

    fig = plt.figure(figsize=(args.figureX, args.figureY), dpi=80, tight_layout=True)
    ax_std = fig.add_subplot(221)
    ax_log = fig.add_subplot(222)
    ax_cdf = fig.add_subplot(223)
    ax_tab = fig.add_subplot(224)

    data_plt = plotGraph(args, PLOT_TYPE_STD, obs_align, mod_align, ['streamflow'], 
                         min_x, max_x, ax_std, secondary='precip', plotColor=args.color)
    fig.text(0.0, 1.0, '(a)')
    plotGraph(args, PLOT_TYPE_LOGY, obs_align, mod_align, ['streamflow'], 
              min_x, max_x, ax_log, plotColor=args.color)
    fig.text(1.0, 1.0, '(b)')
    plotGraph(args, PLOT_TYPE_CDF, obs_align, mod_align, ['streamflow'], 
              min_x, max_x, ax_cdf, plotColor=args.color)
    fig.text(0.0, 0.5, '(c)')
    
    col_names = ['Observed', 'Modeled']
    plotTable(args, col_names, obs_align, mod_align, ax_tab)
    fig.text(1.0, 0.5, '(d)')

    # Figure annotations
    if args.title:
        fig.suptitle(args.title, y=1.01)

    legend_items = ['Observed', 'Modeled']
    fig.legend( data_plt, legend_items, 'lower right', fontsize=10, ncol=2 )

    # Output plot
    if args.color:
        outname = "%s_color" % (args.outname,)
    else:
        outname=args.outname
    plot_filename_png = "%s.png" % (outname,)
    plot_filename_pdf = "%s.pdf" % (outname,)
    plt.savefig(plot_filename_png, bbox_inches='tight') #, bbox_extra_artists=[table])
    plt.savefig(plot_filename_pdf, bbox_inches='tight') #, bbox_extra_artists=[table])



#!/usr/bin/env python
"""@package CreateFlowtable

@brief Create RHESSys flowtable.

This software is provided free of charge under the New BSD License. Please see
the following license information:

Copyright (c) 2013, University of North Carolina at Chapel Hill
All rights reserved.

Redistribution and use in source and binary forms, with or without
modification, are permitted provided that the following conditions are met:
    * Redistributions of source code must retain the above copyright
      notice, this list of conditions and the following disclaimer.
    * Redistributions in binary form must reproduce the above copyright
      notice, this list of conditions and the following disclaimer in the
      documentation and/or other materials provided with the distribution.
    * Neither the name of the University of North Carolina at Chapel Hill nor the
      names of its contributors may be used to endorse or promote products
      derived from this software without specific prior written permission.

THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND
ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
DISCLAIMED. IN NO EVENT SHALL THE UNIVERSITY OF NORTH CAROLINA AT CHAPEL HILL
BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR 
CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE
GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT 
LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT
OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.


@author Brian Miles <brian_miles@unc.edu>


Pre conditions
--------------
1. Configuration file must define the following sections and values:
   'RHESSYS', 'PATH_OF_PARAMDB'

2. The following metadata entry(ies) must be present in the study area section of the metadata associated with the project directory:
   dem_res_x
   dem_res_y

2. The following metadata entry(ies) must be present in the RHESSys section of the metadata associated with the project directory:
   grass_dbase
   grass_location
   grass_mapset
   rhessys_dir
   cf_bin
   worldfile_zero
   template

3. The following metadata entry(ies) must be present in the GRASS section of the metadata associated with the project directory:
   dem_rast
   basin_rast
   slope_rast
   streams_rast
   zero_rast
   roads_rast [optional]
   roof_connectivity_rast [optional]
   impervious_rast [optional]
   
4. The following metadata entry(ies) will be used if present in the GRASS section of the metadata associated with the project directory:
   stream_burned_dem_rast
   
Post conditions
---------------
1. Flowtable will be created in the RHESSys folder of the project directory.

2. File named "cf.out" will be created in the flowtable folder of the RHESSys folder.

3. Will write the following entry(ies) to the RHESSys section of metadata associated with the project directory:
   surface_flowtable
   subsurface_flowtable
   flowtable_cmd

Usage:
@code
CreateFlowtable.py -p /path/to/project_dir
@endcode

@note EcoHydroWorkflowLib configuration file must be specified by environmental variable 'ECOHYDROWORKFLOW_CFG',
or -i option must be specified. 
"""
import argparse

from ecohydrolib.grasslib import *

from rhessysworkflows.context import Context
from rhessysworkflows.metadata import RHESSysMetadata
from rhessysworkflows.rhessys import RHESSysPaths

# Handle command line options
parser = argparse.ArgumentParser(description='Create RHESSys flowtable using GRASS GIS data and createflowpaths utility')
parser.add_argument('-i', '--configfile', dest='configfile', required=False,
                    help='The configuration file. Must define section "GRASS" and option "GISBASE"')
parser.add_argument('-p', '--projectDir', dest='projectDir', required=True,
                    help='The directory to which metadata, intermediate, and final files should be saved')
parser.add_argument('--routeRoads', dest='routeRoads', required=False, action='store_true',
                    help='Tell createflowpaths to route flow from roads to the nearest stream pixel (requires roads_rast to be defined in metadata)')
parser.add_argument('--routeRoofs', dest='routeRoofs', required=False, action='store_true',
                    help='Tell createflowpaths to route flow from roof tops based on roof top connectivity to nearest impervious surface (requires roof_connectivity_rast and impervious_rast to be defined in metadata)')
parser.add_argument('-f', '--force', dest='force', action='store_true',
                    help='Run createflowpaths even if DEM x resolution does not match y resolution')
parser.add_argument('--ignoreBurnedDEM', dest='ignoreBurnedDEM', action='store_true', required=False,
                    help='Ignore stream burned DEM, if present. Default DEM raster will be used for all operations. If not specified and if stream burned raster is present, stream burned DEM will be used for generating the flow table.')
parser.add_argument('-v', '--verbose', dest='verbose', action='store_true',
                    help='Print detailed information about what the program is doing')
args = parser.parse_args()
cmdline = RHESSysMetadata.getCommandLine()

configFile = None
if args.configfile:
    configFile = args.configfile

context = Context(args.projectDir, configFile) 

# Check for necessary information in metadata
studyArea = RHESSysMetadata.readStudyAreaEntries(context)
grassMetadata = RHESSysMetadata.readGRASSEntries(context)
metadata = RHESSysMetadata.readRHESSysEntries(context)

if not 'dem_res_x' in studyArea:
    sys.exit("Metadata in project directory %s does not contain a DEM x resolution" % (context.projectDir,))
if not 'dem_res_y' in studyArea:
    sys.exit("Metadata in project directory %s does not contain a DEM y resolution" % (context.projectDir,))

if not 'grass_dbase' in metadata:
    sys.exit("Metadata in project directory %s does not contain a GRASS Dbase" % (context.projectDir,)) 
if not 'grass_location' in metadata:
    sys.exit("Metadata in project directory %s does not contain a GRASS location" % (context.projectDir,)) 
if not 'grass_mapset' in metadata:
    sys.exit("Metadata in project directory %s does not contain a GRASS mapset" % (context.projectDir,))
if not 'rhessys_dir' in metadata:
    sys.exit("Metadata in project directory %s does not contain a RHESSys directory" % (context.projectDir,))
if not 'cf_bin' in metadata:
    sys.exit("Metadata in project directory %s does not contain a createflowpaths executable" % (context.projectDir,))
if not 'worldfile_zero' in metadata:
    sys.exit("Metadata in project directory %s does not contain an initial worldfile" % (context.projectDir,))
if not 'template' in metadata:
    sys.exit("Metadata in project directory %s does not contain a template" % (context.projectDir,))
if not 'basin_rast' in grassMetadata:
    sys.exit("Metadata in project directory %s does not contain a GRASS dataset with a basin raster" % (context.projectDir,))    
if not 'dem_rast' in grassMetadata:
    sys.exit("Metadata in project directory %s does not contain a GRASS dataset with a DEM raster" % (context.projectDir,))
if not 'slope_rast' in grassMetadata:
    sys.exit("Metadata in project directory %s does not contain a GRASS dataset with a slope raster" % (context.projectDir,))
if not 'streams_rast' in grassMetadata:
    sys.exit("Metadata in project directory %s does not contain a GRASS dataset with a stream raster" % (context.projectDir,))
if not 'zero_rast' in grassMetadata:
    sys.exit("Metadata in project directory %s does not contain a GRASS dataset with a zero raster" % (context.projectDir,))

if args.routeRoads:
    if not 'roads_rast' in grassMetadata:
        sys.exit("Metadata in project directory %s does not contain a GRASS dataset with a roads raster" % (context.projectDir,))

if args.routeRoofs:
    if not 'roof_connectivity_rast' in grassMetadata:
        sys.exit("Metadata in project directory %s does not contain a GRASS dataset with a roofs raster" % (context.projectDir,))
    if not 'impervious_rast' in grassMetadata:
        sys.exit("Metadata in project directory %s does not contain a GRASS dataset with a impervious raster" % (context.projectDir,))

demResX = float(studyArea['dem_res_x'])
demResY = float(studyArea['dem_res_y'])
if demResX != demResY:
    sys.stdout.write("DEM x resolution (%f) does not match y resolution (%f)" %
                     (demResX, demResY) )
    if not args.force:
        sys.exit('Exiting.  Use --force option to override')

# Determine DEM raster to use
demRast = grassMetadata['dem_rast']
if ('stream_burned_dem_rast' in grassMetadata) and (not args.ignoreBurnedDEM):
    demRast = grassMetadata['stream_burned_dem_rast']
sys.stdout.write("Using raster named '%s' to calculate flow direction map\n" % (demRast,) )
    
rhessysDir = metadata['rhessys_dir']
paths = RHESSysPaths(args.projectDir, rhessysDir)

# Set up GRASS environment
grassDbase = os.path.join(context.projectDir, metadata['grass_dbase'])
grassConfig = GRASSConfig(context, grassDbase, metadata['grass_location'], metadata['grass_mapset'])
grassLib = GRASSLib(grassConfig=grassConfig)

# Make sure mask and region are properly set
result = grassLib.script.run_command('g.region', rast=demRast)
if result != 0:
    sys.exit("g.region failed to set region to DEM, returning %s" % (result,))

basinRast = grassMetadata['basin_rast']
result = grassLib.script.run_command('r.mask', flags='o', input=basinRast, maskcats='1')
if result != 0:
    sys.exit("r.mask failed to set mask to basin, returning %s" % (result,))

cfPath = os.path.join(context.projectDir, metadata['cf_bin'])
templatePath = os.path.join(context.projectDir, metadata['template'])
if args.verbose:
    print(templatePath)
# We don't want the '_init' in the flow table name
flowTableNameBase = os.path.basename(metadata['worldfile_zero']).split('_')[0]
flowOutpath = os.path.join(paths.RHESSYS_FLOW, flowTableNameBase)
cfOutpath = os.path.join(paths.RHESSYS_FLOW, 'cf.out')
if args.verbose:
    print(flowOutpath)

roads = None
if args.routeRoads:
    roads = grassMetadata['roads_rast']
else:
    roads = grassMetadata['zero_rast']

roofs = None
impervious = None
if args.routeRoofs:
    roofs = grassMetadata['roof_connectivity_rast']
    impervious = grassMetadata['impervious_rast']
    surfaceFlowtable = "%s_surface.flow" % (flowTableNameBase, )
    subsurfaceFlowtable = "%s_subsurface.flow" % (flowTableNameBase, )
else:
    surfaceFlowtable = subsurfaceFlowtable = "%s.flow" % (flowTableNameBase, )
  
# Run CF
sys.stdout.write('Running createflowpaths (this may take a few minutes)...')
sys.stdout.flush()
p = grassLib.script.pipe_command(cfPath, out=flowOutpath, template=templatePath,
                                 dem=demRast, 
                                 slope=grassMetadata['slope_rast'],
                                 stream=grassMetadata['streams_rast'],
                                 road=roads, roof=roofs, impervious=impervious,
                                 cellsize=demResX)
(pStdout, pStderr) = p.communicate()

if args.verbose:
    print("CF output:")
    print(pStdout)
    if pStderr:
        print(pStderr)

if p.returncode != 0:
    sys.exit("createflowpaths failed, returning %s" % ( str(p.returncode),))

# Write cf output to project directory
cfOut = open(cfOutpath, 'w')
cfOut.write(pStdout)
if pStderr:
    cfOut.write("\n\nStandard error output:\n\n")
    cfOut.write(pStderr)
cfOut.close()

cfCmd = "%s out=%s template=%s dem=%s slope=%s stream=%s road=%s roof=%s impervious=%s cellsize=%s" % \
    (cfPath, flowOutpath, templatePath, demRast, grassMetadata['slope_rast'],
     grassMetadata['streams_rast'], roads, roofs, impervious, demResX)
RHESSysMetadata.writeRHESSysEntry(context, 'flowtable_cmd', cfCmd)
RHESSysMetadata.writeRHESSysEntry(context, 'surface_flowtable', paths.relpath(os.path.join(paths.RHESSYS_FLOW, surfaceFlowtable) ) )
RHESSysMetadata.writeRHESSysEntry(context, 'subsurface_flowtable', paths.relpath(os.path.join(paths.RHESSYS_FLOW, subsurfaceFlowtable) ) )

sys.stdout.write('\nFinished creating flowtable\n')

# Write processing history
RHESSysMetadata.appendProcessingHistoryItem(context, cmdline)


"""@package rhessysworkflows.command.modelrun
    
@brief RHESSysWorkflows commands related to running RHESSys models

This software is provided free of charge under the New BSD License. Please see
the following license information:

Copyright (c) 2015, University of North Carolina at Chapel Hill
All rights reserved.

Redistribution and use in source and binary forms, with or without
modification, are permitted provided that the following conditions are met:
    * Redistributions of source code must retain the above copyright
      notice, this list of conditions and the following disclaimer.
    * Redistributions in binary form must reproduce the above copyright
      notice, this list of conditions and the following disclaimer in the
      documentation and/or other materials provided with the distribution.
    * Neither the name of the University of North Carolina at Chapel Hill nor the
      names of its contributors may be used to endorse or promote products
      derived from this software without specific prior written permission.

THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND
ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
DISCLAIMED. IN NO EVENT SHALL THE UNIVERSITY OF NORTH CAROLINA AT CHAPEL HILL
BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR 
CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE
GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT 
LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT
OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.


@author Brian Miles <brian_miles@unc.edu>
"""
import os
import sys
from subprocess import *
import datetime
import shutil

from rhessysworkflows.command.base import GrassCommand
from rhessysworkflows.command.exceptions import MetadataException
from rhessysworkflows.command.exceptions import RunException

from rhessysworkflows.rhessys import RHESSysPaths
from rhessysworkflows.metadata import RHESSysMetadata
from rhessysworkflows.rhessys import datetimeToString
from rhessysworkflows.rhessys import generateCommandString
from rhessysworkflows.worldfileio import getClimateBaseStationFilenames
from rhessysworkflows.climateio import getStartAndEndDateForClimateStation

class LAIReadMultiple(GrassCommand):
    def __init__(self, projectDir, configFile=None, outfp=sys.stdout):
        """ Construct a LAIReadMultiple command.
        Arguments:
        projectDir -- string    The path to the project directory
        configFile -- string    The path to an EcohydroLib configuration file
        outfp -- file-like object    Where output should be written to
        
        """
        super(LAIReadMultiple, self).__init__(projectDir, configFile, outfp)
        
    def checkMetadata(self, *args, **kwargs):
        """ Check to make sure the project directory has the necessary metadata to run this command.
        """
        super(LAIReadMultiple, self).checkMetadata()
        
        # Check for necessary information in self.metadata
        if not 'subbasin_masks' in self.metadata:
            raise MetadataException("Metadata in project directory %s does not contain multiple worldfile masks" % (self.context.projectDir,))
        if not 'dem_rast' in self.grassMetadata:
            raise MetadataException("Metadata in project directory %s does not contain a GRASS dataset with a DEM raster" % (self.context.projectDir,))
        if not 'hillslope_rast' in self.grassMetadata:
            raise MetadataException("Metadata in project directory %s does not contain a GRASS dataset with a hillslope raster" % (self.context.projectDir,))  
        if not 'zone_rast' in self.grassMetadata:
            raise MetadataException("Metadata in project directory %s does not contain a GRASS dataset with a zone raster" % (self.context.projectDir,)) 
        if not 'patch_rast' in self.grassMetadata:
            raise MetadataException("Metadata in project directory %s does not contain a GRASS dataset with a patch raster" % (self.context.projectDir,)) 
        if not 'stratum_rast' in self.grassMetadata:
            raise MetadataException("Metadata in project directory %s does not contain a GRASS dataset with a stratum raster" % (self.context.projectDir,)) 
        if not 'lai_rast' in self.grassMetadata:
            raise MetadataException("Metadata in project directory %s does not contain a GRASS dataset with an LAI raster" % (self.context.projectDir,)) 
        
        if not 'rhessys_dir' in self.metadata:
            raise MetadataException("Metadata in project directory %s does not contain a RHESSys directory" % (self.context.projectDir,))
        if not 'rhessys_bin' in self.metadata:
            raise MetadataException("Metadata in project directory %s does not contain a RHESSys binary" % (self.context.projectDir,))
        if not 'lairead_bin' in self.metadata:
            raise MetadataException("Metadata in project directory %s does not contain a lairead executable" % (self.context.projectDir,))
        if not 'worldfiles_init' in self.metadata:
            raise MetadataException("Metadata in project directory %s does not contain initial worldfiles" % (self.context.projectDir,))
        if not 'allometric_table' in self.metadata:
            raise MetadataException("Metadata in project directory %s does not contain an allometric table" % (self.context.projectDir,))
        
        topmodel = kwargs.get('topmodel', False)
        if not topmodel:
            if not 'surface_flowtables' in self.metadata:
                raise MetadataException("Metadata in project directory %s does not contain surface flowtables" % (self.context.projectDir,))
            if not 'subsurface_flowtables' in self.metadata:
                raise MetadataException("Metadata in project directory %s does not contain a subsurface flowtables" % (self.context.projectDir,))
        
    def run(self, *args, **kwargs):
        """ Run lairead for multiple worldfiles 
        
        Arguments:
        topmodel -- boolean   Whether to run RHESSys in TOPMODEL model. Default: False.
        verbose -- boolean    Produce verbose output. Default: False.
        """
        verbose = kwargs.get('verbose', False)
        topmodel = kwargs.get('topmodel', False)
        
        self.checkMetadata(topmodel=topmodel)
        
        rhessysDir = self.metadata['rhessys_dir']
        self.paths = RHESSysPaths(self.context.projectDir, rhessysDir)
        rhessysBinPath = os.path.join(self.context.projectDir, self.metadata['rhessys_bin'])
        
        # Make sure region is properly set
        demRast = self.grassMetadata['dem_rast']
        result = self.grassLib.script.run_command('g.region', rast=demRast)
        if result != 0:
            raise RunException("g.region failed to set region to DEM, returning {0}".format(result))
        
        # Run lairead for each worldfile
        lairead_tecfiles = []
        final_worldfiles = []
        worldfiles = self.metadata['worldfiles_init'].split(RHESSysMetadata.VALUE_DELIM)
        masks = self.metadata['subbasin_masks'].split(RHESSysMetadata.VALUE_DELIM)
        
        surfaceFlowtables = subsurfaceFlowtables = None
        if not topmodel:
            surfaceFlowtables = self.metadata['surface_flowtables'].split(RHESSysMetadata.VALUE_DELIM)
            subsurfaceFlowtables = self.metadata['subsurface_flowtables'].split(RHESSysMetadata.VALUE_DELIM)
        
        for (i, worldfile) in enumerate(worldfiles):
            worldfilePath = os.path.join(self.context.projectDir, worldfile)
            worldfileDir = os.path.dirname(worldfilePath)
            
            surfaceFlowtable = subsurfaceFlowtable = None
            if not topmodel:
                surfaceFlowtable = surfaceFlowtables[i] # Assumption: worldfiles and flowtables lists are in the same order
                subsurfaceFlowtable = subsurfaceFlowtables[i] # Assumption: worldfiles and flowtables lists are in the same order
                
            # Mask to correct mask
            mask = masks[i] # Assumption: worldfiles and masks lists are in the same order
            result = self.grassLib.script.run_command('r.mask', flags='o', input=mask, maskcats='1',
                                                      quiet=True)
            if result != 0:
                raise RunException("r.mask failed to set mask to sub-basin {0}, returning {1}".format(mask,
                                                                                                      result))
            ## 1. Determine legal simulation start and date from climate data 
            # Read first climate station from worldfile
            header = "{0}.hdr".format(worldfile)
            headerPath = os.path.join(self.context.projectDir, header)
            stations = getClimateBaseStationFilenames(headerPath)
            if not len(stations) > 0:
                raise RunException("No climate stations found in worldfile header {0}".format(headerPath))
            firstStationPath = os.path.normpath( os.path.join(self.paths.RHESSYS_DIR, stations[0]) )
            if verbose:
                self.outfp.write("First climate station in worldfile: %s\n" % (firstStationPath,) )
            
            # Read climate timeseries for start and end date, write to metadata
            (startDate, endDate) = getStartAndEndDateForClimateStation(firstStationPath, self.paths)
            if verbose:
                self.outfp.write("start date: %s, end date: %s\n" % ( str(startDate), str(endDate) ) )
            fourDays = datetime.timedelta(days=4)
            if endDate - startDate < fourDays:
                raise RunException("Climate time-series defined by station %s is too short to run lairead (less than four-days long)" %
                                   (firstStationPath,) )
            
            ## 2. Run LAI read to generate redefine worldfile
            tecDurRedef = datetime.timedelta(days=1)
            tecRedef = startDate + tecDurRedef
            laireadPath = os.path.join(self.context.projectDir, self.metadata['lairead_bin'])
            oldWorldPath = os.path.join(self.context.projectDir, worldfile)
            redefWorldName = "%s.Y%dM%dD%dH%d" % \
                (worldfile, tecRedef.year, tecRedef.month, tecRedef.day, tecRedef.hour)
            redefWorldPath = os.path.join(self.context.projectDir, redefWorldName)
            allomPath = os.path.join(self.context.projectDir, self.metadata['allometric_table'])
            
            if verbose:
                self.outfp.write("\nRunning lairead for subbasin {0}...".format(mask))
                
            p = self.grassLib.script.pipe_command(laireadPath, old=oldWorldPath, redef=redefWorldPath,
                                                  allom=allomPath, lai=self.grassMetadata['lai_rast'],
                                                  vegid=self.grassMetadata['stratum_rast'],
                                                  zone=self.grassMetadata['zone_rast'],
                                                  hill=self.grassMetadata['hillslope_rast'],
                                                  patch=self.grassMetadata['patch_rast'],
                                                  mask=mask)
            (stdoutStr, stderrStr) = p.communicate() 
            result = p.returncode
            if result != 0:
                self.outfp.write(stdoutStr)
                raise RunException("\nlairead failed, returning %s" % (result,))
            
            if verbose:
                self.outfp.write('done\n')
                        
            ## 3. Write TEC file for redefining the initial flow table
            ##    Redefine on the second day of the simulation, write output
            ##    on the third day
            tecName = "tec.lairead_{0}".format(mask)
            tecPath = os.path.join(self.paths.RHESSYS_TEC, tecName)
            tecDurOutput = datetime.timedelta(days=2)
            tecOutput = startDate + tecDurOutput
            
            f = open(tecPath, 'w')
            f.write("%s redefine_world%s" % 
                    (datetimeToString(tecRedef), os.linesep) )
            f.write("%s output_current_state%s" %
                    (datetimeToString(tecOutput), os.linesep) )
            f.close()
            lairead_tecfiles.append(tecPath)
            
            ## 4. Run RHESSys for the first 4 legal days with redefine TEC
            rhessysStart = startDate
            rhessysDur = datetime.timedelta(days=3)
            rhessysEnd = startDate + rhessysDur
            surfaceFlowtablePath = subSurfaceFlowtablePath = None
            if not topmodel:
                surfaceFlowtablePath = os.path.join(self.context.projectDir, surfaceFlowtable)
                subsurfaceFlowtablePath = os.path.join(self.context.projectDir, subsurfaceFlowtable)
            
            rhessysCmd = generateCommandString(rhessysBinPath, None,
                                               rhessysStart, rhessysEnd,
                                               tecPath, oldWorldPath,
                                               surfaceFlowtablePath, subsurfaceFlowtablePath)
            if verbose:
                self.outfp.write('\nRunning RHESSys to redefine worldfile with vegetation carbon stores...\n')
                self.outfp.write(rhessysCmd)
                self.outfp.write('\n')
            
            cmdArgs = rhessysCmd.split()
            process = Popen(cmdArgs, cwd=self.paths.RHESSYS_DIR, stdout=PIPE, stderr=PIPE)
            (process_stdout, process_stderr) = process.communicate()
            if verbose:
                self.outfp.write(process_stdout)
                self.outfp.write(process_stderr)
            if process.returncode != 0:
                raise RunException("\n\nRHESSys failed, returning %s" % (process.returncode,) )
            
            if verbose:    
                sys.stdout.write('done\n')
                
            ## 5. Rename redefine worldfile, write to metadata
            outputWorldName = "%s.Y%dM%dD%dH%d.state" % \
                (worldfile, tecOutput.year, tecOutput.month, tecOutput.day, tecOutput.hour)
            outputWorldPath = os.path.join(self.context.projectDir, outputWorldName)
            
            if not os.access(outputWorldPath, os.W_OK):
                raise RunException("Unable to find redefined worldfile %s" % (outputWorldPath,) )
                
            newWorldName = "world_{0}".format(mask)
            newWorldPath = os.path.join(self.paths.RHESSYS_WORLD, newWorldName)
            
            shutil.move(outputWorldPath, newWorldPath)
            if not os.path.exists(newWorldPath):
                raise RunException("Failed to copy redefined worldfile %s to %s" % (outputWorldPath, newWorldPath) )
            
            final_worldfiles.append(newWorldPath)
            
            # Copy world file header from init worldfile to final world file
            newHeader = "%s.hdr" % (newWorldName,)
            newHeaderPath = os.path.join(self.paths.RHESSYS_WORLD, newHeader)
            shutil.copyfile(headerPath, newHeaderPath)
            
        if verbose:    
            sys.stdout.write('\n\nSuccessfully used lairead to initialize vegetation carbon stores.\n')
        
        # Write metadata    
        RHESSysMetadata.writeRHESSysEntry(self.context, 'worldfiles',
                                          RHESSysMetadata.VALUE_DELIM.join([self.paths.relpath(w) for w in final_worldfiles]))
        RHESSysMetadata.writeRHESSysEntry(self.context, 'lairead_tecfiles', 
                                          RHESSysMetadata.VALUE_DELIM.join([self.paths.relpath(t) for t in lairead_tecfiles]))
        RHESSysMetadata.writeRHESSysEntry(self.context, 'lairead_mode_topmodel', str(topmodel))
            
        # Write processing history
        RHESSysMetadata.appendProcessingHistoryItem(self.context, RHESSysMetadata.getCommandLine())


"""@package rhessysworkflows.ginotebook

@brief Tools that enable interaction with GI Database:
https://github.com/ResearchSoftwareInstitute/ginotebook

This software is provided free of charge under the New BSD License. Please see
the following license information:

Copyright (c) 2016, University of North Carolina at Chapel Hill
All rights reserved.

Redistribution and use in source and binary forms, with or without
modification, are permitted provided that the following conditions are met:
    * Redistributions of source code must retain the above copyright
      notice, this list of conditions and the following disclaimer.
    * Redistributions in binary form must reproduce the above copyright
      notice, this list of conditions and the following disclaimer in the
      documentation and/or other materials provided with the distribution.
    * Neither the name of the University of North Carolina at Chapel Hill nor the
      names of its contributors may be used to endorse or promote products
      derived from this software without specific prior written permission.

THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND
ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
DISCLAIMED. IN NO EVENT SHALL THE UNIVERSITY OF NORTH CAROLINA AT CHAPEL HILL
BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE
GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT
OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.


@author Brian Miles <brian_miles@unc.edu>
"""
import json
from collections import OrderedDict

import requests

from .compat import http_responses


DEFAULT_HOSTNAME = 'gidesigner.renci.org'
DEFAULT_API_ROOT = 'ginotebook/api'

GI_TYPE_RAIN_GARDEN = 'Rain Garden'
GI_TYPE_TREE = 'Tree'
GI_TYPE_GREEN_ROOF = 'Green roof'
GI_TYPES = (GI_TYPE_RAIN_GARDEN, GI_TYPE_TREE, GI_TYPE_GREEN_ROOF)


class GINotebookException(Exception):
    def __init__(self, args):
        super(GINotebookException, self).__init__(args)


class GINotebookHTTPException(GINotebookException):
    """ Exception used to communicate HTTP errors from GI Notebook server
        Arguments in tuple passed to constructor must be: (url, status_code, params).
        url and status_code are of type string, while the optional params argument
        should be a dict.
    """
    def __init__(self, args):
        super(GINotebookHTTPException, self).__init__(args)
        self.url = args[0]
        self.method = args[1]
        self.status_code = args[2]
        if len(args) >= 4:
            self.params = args[3]
        else:
            self.params = None

    def __str__(self):
        msg = "Received status {status_code} {status_msg} when accessing {url} " + \
              "with method {method} and params {params}."
        return msg.format(status_code=self.status_code,
                          status_msg=http_responses[self.status_code],
                          url=self.url,
                          method=self.method,
                          params=self.params)

    def __unicode__(self):
        return unicode(str(self))


class GINotebookIntegrityError(GINotebookException):
    def __init__(self, args):
        super(GINotebookIntegrityError, self).__init__(args)
        self.obj_url1 = args[0]
        self.obj_url2 = args[1]
        self.field = args[2]

    def __str__(self):
        msg = "Expected {obj1} but got {obj2} for {field}."
        return msg.format(obj1=self.obj_url1, obj2=self.obj_url2, field=self.field)

    def __unicode__(self):
        return unicode(str(self))


class RHESSysDefaultType(object):
    """ Represents a RHESSys parameter default type (e.g. soil, stratum)
    """
    def __init__(self, id, url, name, rhessys_default_id):
        self.id = id
        self.url = url
        self.name = name
        self.rhessys_default_id = rhessys_default_id


class SoilType(RHESSysDefaultType):
    """ Represents a SoilType as presented by the GI Notebook rhessys_soil_types REST API end point
    """
    def __init__(self, id, url, name, rhessys_default_id):
        super(SoilType, self).__init__(id, url, name, rhessys_default_id)


class StratumType(RHESSysDefaultType):
    """ Represents a StratumType as presented by the GI Notebook rhessys_stratum_types REST API end point
    """
    def __init__(self, id, url, name, rhessys_default_id):
        super(StratumType, self).__init__(id, url, name, rhessys_default_id)


class GIElement(object):
    """ Represents a GI Element as presented by the GI Notebook gi_element REST API end point
    """
    def __init__(self, id, url, name, model_3d_url, model_planview_url, soil_depth, ponding_depth,
                 major_axis, minor_axis,
                 stratum_type=None, soil_type=None):
        self.id = id
        self.url = url
        self.name = name
        self.model_3d_url = model_3d_url
        self.model_planview_url = model_planview_url
        self.soil_depth = soil_depth
        self.ponding_depth = ponding_depth
        self.major_axis = major_axis
        self.minor_axis = minor_axis
        self.stratum_type = stratum_type
        self.soil_type = soil_type

    def get_properties(self, dict_to_use=None, key_prefix=None, shorten=False):
        """ Generate a dict consisting of the properties of a GIElement.

        @param: dict_to_use: The dict to store properties in.  If None, a new dict will be created.
        @param: key_prefix: Prefix to be prepended onto each key name.  Prefix will be separate from the
        rest of the key by "_".
        @param: shorten: Use short names for properties if True.
        @return: Dict of properties.
        """
        if dict_to_use:
            p = dict_to_use
        else:
            p = {}

        temp = OrderedDict()
        temp['id'] = self.id
        temp['url'] = self.url
        temp['name'] = self.name

        if shorten:
            soil_depth = 'pedz'
            ponding_depth = 'pondz'
            major_axis = 'majax'
            minor_axis = 'minax'
            stratum_name = 'vegnm'
            stratum_default_id = 'vegid'
            soil_name = 'pednm'
            soil_default_id = 'pedid'
        else:
            soil_depth = 'soil_depth'
            ponding_depth = 'ponding_depth'
            major_axis = 'major_axis'
            minor_axis = 'minor_axis'
            stratum_name = 'stratum_name'
            stratum_default_id = 'stratum_default_id'
            soil_name = 'soil_name'
            soil_default_id = 'soil_default_id'

        temp[soil_depth] = self.soil_depth
        temp[ponding_depth] = self.ponding_depth
        temp[major_axis] = self.major_axis
        temp[minor_axis] = self.minor_axis
        if self.stratum_type:
            temp[stratum_name] = self.stratum_type.name
            temp[stratum_default_id] = int(self.stratum_type.rhessys_default_id)
        if self.soil_type:
            temp[soil_name] = self.soil_type.name
            temp[soil_default_id] = int(self.soil_type.rhessys_default_id)

        if key_prefix:
            # Re-write values with keys prefixed by key_prefix
            temp = {"{0}_{1}".format(key_prefix, k): v for (k, v) in temp.iteritems()}

        p.update(temp)

        return p


class GITemplate(object):
    """ Represents a GI Template as presented by the GI Notebook gi_template REST API end point
    """
    def __init__(self, id, url, name, gi_type, model_3d_url, model_planview_url):
        self.id = id
        self.url = url
        self.name = name
        self.gi_type = gi_type
        self.model_3d_url = model_3d_url
        self.model_planview_url = model_planview_url
        self.gi_elements = []

    def add_element(self, gi_element):
        """ Associate an instance with this template

        @param gi_element: A GIElement object
        @return:
        """
        self.gi_elements.append(gi_element)

    def get_properties(self, dict_to_use=None, flatten=True, shorten=False):
        """ Generate a dict consisting of the properties of a GIInstance.

        @param: dict_to_use: The dict to store properties in.  If None, a new dict will be created.
        @param: flatten: If False, properties of elements contained in template will be written
        to their own dict with key "element_<id>".  If True, element properties will be stored directly
        in the template properties dict, with "element_<id>_" prepended to each property name.
        @param: shorten: Use short names for properties if True.
        @return: Dict of properties.
        """
        if dict_to_use:
            p = dict_to_use
        else:
            p = {}

        if shorten:
            template_id = 'templt_id'
            template_url = 'templt_url'
            elements = 'elements'
        else:
            template_id = 'template_id'
            template_url = 'template_url'
            elements = 'elements'

        p[template_id] = self.id
        p[template_url] = self.url
        p['type'] = self.gi_type

        if not flatten:
            elems = []
            p[elements] = elems

        for i, element in enumerate(self.gi_elements, 1):
            if flatten:
                if shorten:
                    key = "e_{0}".format(i)
                else:
                    key = "element_{0}".format(i)
                element.get_properties(dict_to_use=p, key_prefix=key, shorten=shorten)
            else:
                elements.append(element.get_properties())

        return p


class GIInstance(object):
    """ Represents a GI Instance as presented by the GI Notebook gi_instance REST API end point
    """
    def __init__(self, id, url, placement_poly, poly_area_sq_meter, scenario=None, template=None):
        """

        @param id:
        @param url:
        @param placement_poly: WGS 84 polygon (a dict representation of the GeoJSON)
        @param poly_area_sq_meter: The area of placement_poly in square meters
        @param scenario: A GIScenario object
        @param template: A GITemplate object
        @return:
        """
        self.id = id
        self.url = url
        self.placement_poly = placement_poly
        self.poly_area_sq_meter = poly_area_sq_meter
        self.scenario = scenario
        self.template = template

    def get_as_geojson_feature(self, flatten=True, shorten=False, filter=None):
        """ Generate a dict representing a GIInstance as a GeoJSON Feature

        @param: flatten If True, compound properties will be flattened into a single namespace such that
        all properties values are simple strings.  This allows GIS clients to easily interpret feature properties.
        If False, properties values may be any JSON object.
        @param: shorten: Use short names for properties if True.
        @param: filter If flatten is True, filter should be a function that takes as input a dictionary
        and returns True if the items in the dictionary pass the filter, False if not.
        @return: Dict representing a GeoJSON Feature.
        """
        if shorten:
            poly_area_sq_meter = 'area_sq_m'
            instance_id = 'inst_id'
            instance_url = 'inst_url'
        else:
            poly_area_sq_meter = 'poly_area_sq_meter'
            instance_id = 'instance_id'
            instance_url = 'instance_url'

        p = OrderedDict([(poly_area_sq_meter, self.poly_area_sq_meter),
                         (instance_id, self.id), (instance_url, self.url)])
        if self.template:
            p = self.template.get_properties(p, flatten=flatten, shorten=shorten)

        feature = None
        filter_pass = True
        if flatten and filter:
            if not filter(p):
                filter_pass = False

        if filter_pass:
            feature = OrderedDict([('type', 'Feature'), ('id', self.id),
                                    ('geometry', self.placement_poly),
                                    ('properties', p)])
        return feature


class GIScenario(object):
    """ Represents a GI Scenario as presented by the GI Notebook gi_scenario REST API end point
    """
    def __init__(self, id, url, name, description, immutable, watershed_url):
        """

        @param id:
        @param url:
        @param name:
        @param description:
        @param immutable:
        @param watershed_url:
        @return:
        """
        self.id = id
        self.url = url
        self.name = name
        self.description = description
        self.immutable = immutable
        self.watershed_url = watershed_url
        self.gi_instances = []

    def add_instance(self, gi_instance):
        """ Associate a GIInstance with this scenario

        @param gi_instance: A GIInstance object
        @return:
        """
        gi_instance.scenario = self
        self.gi_instances.append(gi_instance)

    def get_instances_as_geojson(self, indent=None, flatten=True, shorten=False,
                                 filter=None):
        """ Generate a GeoJSON FeatureCollection representing all GI Instances associated with a scenario.

        @param: indent Integer by which JSON output will be indented for pretty printing
        @param: flatten If True, compound properties will be flattened into a single namespace such that
        all properties values are simple strings.  This allows GIS clients to easily interpret feature properties.
        If False, properties values may be any JSON object.
        @param: shorten Use short names for properties if True.
        @param: filter If flatten is True, filter should be a function that takes as input a dictionary
        and returns True if the items in the dictionary pass the filter, False if not.
        @return: String in GeoJSON format
        """
        features = []
        feature_collection = OrderedDict([('type', 'FeatureCollection'), ('features', features)])
        for instance in self.gi_instances:
            feature = instance.get_as_geojson_feature(flatten=flatten, shorten=shorten, filter=filter)
            if feature is not None:
                features.append(feature)

        return json.dumps(feature_collection, indent=indent)


class GINotebook(object):
    """ Class that allows interaction with GI Database:
        https://github.com/ResearchSoftwareInstitute/ginotebook

    """
    _URL_PROTO_WITHOUT_PORT = "{scheme}://{hostname}/{api_root}"
    _URL_PROTO_WITH_PORT = "{scheme}://{hostname}:{port}/{api_root}"

    def __init__(self, hostname=DEFAULT_HOSTNAME, api_root=DEFAULT_API_ROOT,
                 port=None, use_https=True, verify=True, auth_token=None):
        self.hostname = hostname
        self.verify = verify

        self.session = None
        self.auth_header = {}
        if auth_token:
            self.auth_header['Authorization'] = "Token {token}".format(token=auth_token)

        if use_https:
            self.scheme = 'https'
        else:
            self.scheme = 'http'
        self.use_https = use_https

        if port:
            self.port = int(port)
            if self.port < 0 or self.port > 65535:
                raise GINotebookException("Port number {0} is illegal.".format(self.port))
            self.url_base = self._URL_PROTO_WITH_PORT.format(scheme=self.scheme,
                                                             hostname=self.hostname,
                                                             port=self.port,
                                                             api_root=api_root)
        else:
            self.url_base = self._URL_PROTO_WITHOUT_PORT.format(scheme=self.scheme,
                                                                hostname=self.hostname,
                                                                api_root=api_root)

    def _request(self, method, url, params=None, data=None, files=None, headers=None, stream=False):
        if headers:
            h = dict(headers)
            h.update(self.auth_header)
        else:
            h = dict(self.auth_header)

        r = requests.request(method, url, params=params, data=data, files=files, headers=h, stream=stream,
                             verify=self.verify)
        return r

    def _get_resource(self, endpoint, id=None, url=None):
        if id:
            url = "{url_base}/{endpoint}/{id}/".format(url_base=self.url_base, endpoint=endpoint, id=id)
        elif not url:
            GINotebookException("Resource URL not specified")

        r = self._request('GET', url)
        if r.status_code != 200:
            raise GINotebookHTTPException((url, 'GET', r.status_code))
        return r.json()

    def get_scenario(self, id=None, url=None):
        """ Get GIScenario resource from the GI Notebook

        @param id: The ID of the GIScenario resource to download from the GI Notebook
        @param url: The URL of the GIScenario resource to download from the GI Notebook
        @return: GIScenario instance representing the resource
        """
        raw = self._get_resource('gi_scenarios', id=id, url=url)
        scenario = GIScenario(raw['id'], raw['url'], raw['name'], raw['description'], raw['immutable'],
                              raw['watershed'])
        for instance_url in raw['giinstances']:
            instance = self.get_instance(url=instance_url)
            scenario.add_instance(instance)

        return scenario

    def get_instance(self, id=None, url=None):
        """ Get GIInstance resource from the GI Notebook

        @param id: The ID of the GIInstance resource to download from the GI Notebook
        @param url: The URL of the GIInstance resource to download from the GI Notebook
        @return: GIInstance instance representing the resource
        """
        raw = self._get_resource('gi_instances', id=id, url=url)
        template = self.get_template(url=raw['template'])
        instance = GIInstance(raw['id'], raw['url'], raw['placement_poly'], raw['placement_poly_area_sq_m'],
                              template=template)

        return instance

    def get_template(self, id=None, url=None):
        """ Get GITemplate resource from the GI Notebook

        @param id: The ID of the GITemplate resource to download from the GI Notebook
        @param url: The URL of the GITemplate resource to download from the GI Notebook
        @return: GITemplate instance representing the resource
        """
        raw = self._get_resource('gi_templates', id=id, url=url)
        gi_type = self.get_type(url=raw['gi_type'])
        template = GITemplate(raw['id'], raw['url'], raw['name'], gi_type,
                              raw['model_3d'], raw['model_planview'])
        for element_url in raw['gi_elements']:
            element = self.get_element(url=element_url)
            template.add_element(element)

        return template

    def get_type(self, id=None, url=None):
        """ Get GI type resource from the GI Notebook

        @param id: The ID of the GI type resource to download from the GI Notebook
        @param url: The URL of the GI type resource to download from the GI Notebook
        @return: String representing the GI type
        """
        raw = self._get_resource('gi_types', id=id, url=url)

        return raw['name']

    def get_element(self, id=None, url=None):
        """ Get GIElement resource from the GI Notebook

        @param id: The ID of the GIElement resource to download from the GI Notebook
        @param url: The URL of the GIElement resource to download from the GI Notebook
        @return: GIElement instance representing the resource
        """
        raw = self._get_resource('gi_elements', id=id, url=url)
        stratum_type = None
        if raw['stratum_type']:
            stratum_type = self.get_stratum_type(url=raw['stratum_type'])
        soil_type = None
        if raw['soil_type']:
            soil_type = self.get_soil_type(url=raw['soil_type'])
        element = GIElement(raw['id'], raw['url'], raw['name'],
                            raw['model_3d'], raw['model_planview'],
                            raw['soil_depth'], raw['ponding_depth'],
                            raw['major_axis'], raw['minor_axis'],
                            stratum_type, soil_type)

        return element

    def get_stratum_type(self, id=None, url=None):
        """ Get StratumType resource from the GI Notebook

        @param id: The ID of the StratumType resource to download from the GI Notebook
        @param url: The URL of the StratumType resource to download from the GI Notebook
        @return: StratumType instance representing the resource
        """
        raw = self._get_resource('rhessys_stratum_types', id=id, url=url)
        stratum_type = StratumType(raw['id'], raw['url'], raw['name'],
                                   raw['rhessys_default_id'])

        return stratum_type

    def get_soil_type(self, id=None, url=None):
        """ Get SoilType resource from the GI Notebook

        @param id: The ID of the SoilType resource to download from the GI Notebook
        @param url: The URL of the SoilType resource to download from the GI Notebook
        @return: SoilType instance representing the resource
        """
        raw = self._get_resource('rhessys_stratum_types', id=id, url=url)
        soil_type = SoilType(raw['id'], raw['url'], raw['name'],
                             raw['rhessys_default_id'])

        return soil_type



#!/usr/bin/env python
"""@package ImportRHESSysSource

@brief Import RHESSys source code into project directory; Will import either from the
Git repository hosted on GitHub, from a local copy of the source tree.  The local 
source tree must be root of the RHESSys source tree and contain the following source
sub-directories: cf, g2w, and rhessys.  Will also import RHESSys ParamDB from GitHub.

This software is provided free of charge under the New BSD License. Please see
the following license information:

Copyright (c) 2013, University of North Carolina at Chapel Hill
All rights reserved.

Redistribution and use in source and binary forms, with or without
modification, are permitted provided that the following conditions are met:
    * Redistributions of source code must retain the above copyright
      notice, this list of conditions and the following disclaimer.
    * Redistributions in binary form must reproduce the above copyright
      notice, this list of conditions and the following disclaimer in the
      documentation and/or other materials provided with the distribution.
    * Neither the name of the University of North Carolina at Chapel Hill nor the
      names of its contributors may be used to endorse or promote products
      derived from this software without specific prior written permission.

THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND
ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
DISCLAIMED. IN NO EVENT SHALL THE UNIVERSITY OF NORTH CAROLINA AT CHAPEL HILL
BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR 
CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE
GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT 
LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT
OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.


@author Brian Miles <brian_miles@unc.edu>


Pre conditions
--------------
1. Configuration file must define the following sections and values:
   'GRASS', 'GISBASE'
   'RHESSYS', 'PATH_OF_GIT'
   'RHESSYS', 'PATH_OF_MAKE'

2. The following metadata entry(ies) must be present in the RHESSys section of the 
   metadata associated with the project directory: 
   rhessys_dir
   
Post conditions
---------------
1. Will import/copy source code into src directory of rhessys_dir

2. Will compile g2w (including rat), cf, and rhessys binaries into bin directory of 
   rhessys_dir 

3. Will write the following entry(ies) to the RHESSys section of metadata associated with the project directory:
   rhessys_src
   rhessys_branch or rhessys_tag (if specified)
   rhessys_sha (if ImportRHESSysSource fetched source from Git repository)
   paramdb_src
   paramdb_sha
   paramdb_dir
   paramdb
   template_template
   exec_dir (directory within project directory containing executables)
   rhessys_bin
   cf_bin
   rat_bin
   g2w_bin
   lairead_bin
   allometric_table

Usage:
@code
ImportRHESSysSource.py -p /path/to/project_dir
@endcode
"""
import os, sys, errno, stat
import argparse
import re
import shutil
from subprocess import *
from cStringIO import StringIO

from rhessysworkflows.context import Context
from rhessysworkflows.metadata import RHESSysMetadata
from rhessysworkflows.rhessys import RHESSysPaths

PARAM_DB_REPO_URL = 'https://github.com/RHESSys/ParamDB.git'
paramDBDir = 'ParamDB'
paramDBName = 'params.sqlite'
TEMPLATE_PATH = os.path.join('templates', 'template.template')
ALLOMETRIC_PATH = os.path.join('allometry', 'allometric.txt')

RHESSYS_REPO_URL = 'https://github.com/RHESSys/RHESSys.git'

# Handle command line options
parser = argparse.ArgumentParser(description='Import RHESSys source code into project directory')
parser.add_argument('-i', '--configfile', dest='configfile', required=False,
                    help='The configuration file.')
parser.add_argument('-p', '--projectDir', dest='projectDir', required=True,
                    help='The directory to which metadata, intermediate, and final files should be saved')
parser.add_argument('-s', '--sourceDir', dest='sourceDir', required=False,
                    help='The directory from which RHESys source should be copied. NOTE: will delete any sources already in the project directory')
group = parser.add_mutually_exclusive_group()
group.add_argument('-t', '--tag', dest='tag', required=False, 
                   help='Use source code from the specified tagged version of RHESSys; applies only when code is cloned from Git repository (i.e. -s not specified)')
group.add_argument('-b', '--branch', dest='branch', required=False,
                   help='Use source code from the specified branch of the RHESSys source; applies only when code is cloned from Git repository (i.e. -s not specified)')
group.add_argument('-c', '--commit', dest='commit', required=False,
                   help='Use source code from the specified commit of the RHESSys source; applies only when code is clone from Git repository (i.e. -s not specified)')
parser.add_argument('--overwrite', dest='overwrite', action='store_true', required=False,
                    help='Overwrite existing source code in the project directory; If specified, will delete existing code before importing new code.  If not specified, new code will be added to existing code.')
args = parser.parse_args()
cmdline = RHESSysMetadata.getCommandLine()

configFile = None
if args.configfile:
    configFile = args.configfile

context = Context(args.projectDir, configFile)

gisBase = context.config.get('GRASS', 'GISBASE')
gitPath = context.config.get('RHESSYS', 'PATH_OF_GIT')
makePath = context.config.get('RHESSYS', 'PATH_OF_MAKE')

# Check for necessary information in metadata
metadata = RHESSysMetadata.readRHESSysEntries(context)
if not 'rhessys_dir' in metadata:
    sys.exit("Metadata in project directory %s does not contain a RHESSys directory" % (context.projectDir,))

paths = RHESSysPaths(args.projectDir, metadata['rhessys_dir'])

paramDBPath = os.path.join(paths.DB_DIR, paramDBDir)
    
if args.overwrite:
    sys.stdout.write("Deleting existing paramDB in %s\n" % (paramDBPath,) )
    # Delete any existing source
    try:
        contents = os.listdir(paramDBPath)
        for entry in contents:
            toDelete = os.path.join(paramDBPath, entry)
            if os.path.isdir(toDelete):
                shutil.rmtree(toDelete)
            else:
                os.unlink(toDelete)
    except OSError as e:
        if e.errno != errno.ENOENT:
            raise e
        
    sys.stdout.write("Deleting existing source code in %s\n" % (paths.RHESSYS_SRC,) )
    # Delete any existing source
    try:
        contents = os.listdir(paths.RHESSYS_SRC)
        for entry in contents:
            toDelete = os.path.join(paths.RHESSYS_SRC, entry)
            if os.path.isdir(toDelete):
                shutil.rmtree(toDelete)
            else:
                os.unlink(toDelete)
    except OSError as e:
        if e.errno != errno.ENOENT:
            raise e

## 1. Fetch ParamDB
gitCommand = "%s clone %s %s" % (gitPath, PARAM_DB_REPO_URL, paramDBPath)
sha1Identifier = 'HEAD'
returnCode = os.system(gitCommand)
if returnCode != 0:
    sys.exit("Git command '%s' failed." % (gitCommand, ) )
# Write metadata
RHESSysMetadata.writeRHESSysEntry(context, 'paramdb_src', PARAM_DB_REPO_URL)
# Get SHA1 hash for version of ParamDB we are using, then store the hash in metadata
gitCommand = "%s rev-parse %s" % (gitPath, sha1Identifier)
process = Popen(gitCommand, shell=True, stdout=PIPE, cwd=paramDBPath)
(processStdout, processStderr) = process.communicate()
try:
    sha = processStdout.splitlines()[0]
    RHESSysMetadata.writeRHESSysEntry(context, 'paramdb_sha', sha)
except IndexError:
    pass
# Find ParamDB
contents = os.listdir(paramDBPath)
found = False
for entry in contents:
    if entry == paramDBName:
        found = True
        break
if not found:
    sys.exit("Unable to find parameter database %s in %s" % (paramDBName, paramDBPath) )
paramDB = os.path.join(paramDBPath, paramDBName)

# Make sure there is a template in the imported source
templatePath = os.path.join(paramDBPath, TEMPLATE_PATH)
#print(templatePath)
if not os.path.exists(templatePath):
    sys.exit("Template template not found in ParamDB git clone")

# Make sure there is an allometric.txt in the imported source
allometricPath = os.path.join(paramDBPath, ALLOMETRIC_PATH)
#print(allometricPath)
if not os.path.exists(allometricPath):
    sys.exit("Allometric table not found in ParamDB git clone")

RHESSysMetadata.writeRHESSysEntry(context, 'paramdb_dir', paths.relpath(paramDBPath) )
RHESSysMetadata.writeRHESSysEntry(context, 'paramdb', paths.relpath(paramDB) )
RHESSysMetadata.writeRHESSysEntry(context, 'template_template', paths.relpath(templatePath) )
RHESSysMetadata.writeRHESSysEntry(context, 'allometric_table', paths.relpath(allometricPath) )

## 2. Import code from local disk, or from GitHub            
if args.sourceDir:
    # Import source code from sourceDir
    if not os.access(args.sourceDir, os.R_OK):
        sys.exit("The specified path of the RHESSys source directory, %s, is not readable" % (args.sourceDir,) )
    if not os.path.isdir(args.sourceDir):
        sys.exit("The specified path of the RHESSys source directory, %s, is not a directory" % (args.sourceDir,) )
    
    g2wSrc = os.path.join(args.sourceDir, 'g2w')
    if not os.path.exists(g2wSrc) or not os.path.isdir(g2wSrc):
        sys.exit("The specified path of the RHESSys source directory, %s, does not seem to contain grass2world source code in a directory called 'g2w'" %
                 (args.sourceDir,) )
    
    cfSrc = os.path.join(args.sourceDir, 'cf')
    if not os.path.exists(cfSrc) or not os.path.isdir(cfSrc):
        sys.exit("The specified path of the RHESSys source directory, %s, does not seem to contain createflowpaths source code in a directory called 'cf'" %
                 (args.sourceDir,) )
    
    modelSrc = os.path.join(args.sourceDir, 'rhessys')
    if not os.path.exists(modelSrc) or not os.path.isdir(modelSrc):
        sys.exit("The specified path of the RHESSys source directory, %s, does not seem to contain RHESSys model source code in a directory called 'rhessys'" %
                 (args.sourceDir,) )
    
    # Delete paths.RHESSYS_SRC so that we can use shutil.copytree (which will recreate paths.RHESSYS_SRC)
    sys.stdout.write("Copying RHESSys source from %s..." % (args.sourceDir,) )
    sys.stdout.flush()
    shutil.rmtree(paths.RHESSYS_SRC)
    shutil.copytree(args.sourceDir, paths.RHESSYS_SRC)
    sys.stdout.write('done\n')
    # Write metadata
    RHESSysMetadata.writeRHESSysEntry(context, 'rhessys_src', os.path.abspath(args.sourceDir) )
    # Remove information about RHESSys if previously fetched from the Git repository 
    RHESSysMetadata.deleteRHESSysEntry(context, 'rhessys_branch')
    RHESSysMetadata.deleteRHESSysEntry(context, 'rhessys_tag')
    RHESSysMetadata.deleteRHESSysEntry(context, 'rhessys_sha')
    
else:   
    # Import from GitHub
    gitCommand = "%s clone %s %s" % (gitPath, RHESSYS_REPO_URL, paths.RHESSYS_SRC)
    sha1Identifier = 'HEAD'
    returnCode = os.system(gitCommand)
    if returnCode != 0:
        sys.exit("Git command '%s' failed." % (gitCommand, ) )
    # Write metadata
    RHESSysMetadata.writeRHESSysEntry(context, 'rhessys_src', RHESSYS_REPO_URL)
    
    if not args.branch and not args.tag and not args.commit:
        args.tag = 'RHESSys-5.18.2' # Default to latest stable version
    
    if args.branch or args.tag or args.commit:
        gitCommand = None
        gitSHA1Command = None
        # Check out desired branch or tag
        if args.branch:
            branchRegex = re.compile('^\s*origin/(.+)$')
            # Lookup available branches
            branches = []
            gitCommand = "%s branch -r" % (gitPath,)
            process = Popen(gitCommand, shell=True, stdout=PIPE, cwd=paths.RHESSYS_SRC)
            (processStdout, processStderr) = process.communicate()
            for line in StringIO(processStdout).readlines():
                m = branchRegex.match(line)
                if m: branches.append(m.group(1))
            
            if not args.branch in branches:
                sys.stdout.write("Branch %s not found, using default branch\n" % (args.branch,) )
            else:
                gitCommand = "%s checkout %s" % (gitPath, args.branch)
                sha1Identifier = args.branch
            # Write metadata
            RHESSysMetadata.writeRHESSysEntry(context, 'rhessys_branch', args.branch)
        if args.tag:
            # Lookup available branches
            tags = []
            gitCommand = "%s tag -l" % (gitPath,)
            process = Popen(gitCommand, shell=True, stdout=PIPE, cwd=paths.RHESSYS_SRC)
            (processStdout, processStderr) = process.communicate()
            for line in StringIO(processStdout).readlines():
                tags.append(line.strip())
            
            if not args.tag in tags:
                sys.stdout.write("Tag %s not found, using default tag\n" % (args.tag,) )
            else:
                gitCommand = "%s checkout %s" % (gitPath, args.tag)
                sha1Identifier = args.tag
            # Write metadata
            RHESSysMetadata.writeRHESSysEntry(context, 'rhessys_tag', args.tag)
        if args.commit:
            sys.stdout.write("Checking out commit {0}\n".format(args.commit))
            gitCommand = "%s checkout %s" % (gitPath, args.commit)
            sha1Identifier = args.commit
        if gitCommand:        
            # Check out the branch or tag
            process = Popen(gitCommand, shell=True, cwd=paths.RHESSYS_SRC)
            returnCode = process.wait()
            if returnCode != 0:
                sys.exit("Git command '%s' failed" % (gitCommand,) )
    # Get SHA1 hash for version of RHESSys we are using, then store the hash in metadata
    if sha1Identifier:
        gitCommand = "%s rev-parse %s" % (gitPath, sha1Identifier)
        process = Popen(gitCommand, shell=True, stdout=PIPE, cwd=paths.RHESSYS_SRC)
        (processStdout, processStderr) = process.communicate()
        try:
            sha = processStdout.splitlines()[0]
            RHESSysMetadata.writeRHESSysEntry(context, 'rhessys_sha', sha)
        except IndexError:
            pass
            
## 3. Compile code
# Set GISBASE (needed to compile g2w, cf, lairead)
os.environ['GISBASE'] = gisBase
permissions = stat.S_IRUSR | stat.S_IWUSR | stat.S_IXUSR
# Build g2w
buildPath = os.path.join(paths.RHESSYS_SRC, 'g2w')
makeCommand = makePath
process = Popen(makeCommand, shell=True, cwd=buildPath)
returnCode = process.wait()
if returnCode != 0:
    sys.exit("Failed to build g2w, return code: %s" % (returnCode,) )
# Copy binaries to paths.RHESSYS_BIN
g2wSrc = os.path.join(buildPath, 'grass2world', 'g2w')
g2wDest = os.path.join(paths.RHESSYS_BIN, 'g2w')
shutil.copyfile(g2wSrc, g2wDest)
os.chmod(g2wDest, permissions)
# Write metadata
RHESSysMetadata.writeRHESSysEntry(context, 'g2w_bin', paths.relpath(g2wDest) )

ratSrc = os.path.join(buildPath, 'AverageTables_Unix', 'rat')
ratDest = os.path.join(paths.RHESSYS_BIN, 'rat')
shutil.copyfile(ratSrc, ratDest)
os.chmod(ratDest, permissions)
# Write metadata
RHESSysMetadata.writeRHESSysEntry(context, 'rat_bin', paths.relpath(ratDest) )

# Build cf
buildPath = os.path.join(paths.RHESSYS_SRC, 'cf')
makeCommand = makePath
process = Popen(makeCommand, shell=True, cwd=buildPath)
returnCode = process.wait()
if returnCode != 0:
    sys.exit("Failed to build cf, return code: %s" % (returnCode,) )

# Find the cf binary
cfBinRegex = re.compile('^cf.+$')
contents = os.listdir(buildPath)
cfBin = None
for entry in contents:
    m = cfBinRegex.match(entry)
    if m:
        cfBin = entry
        break
assert(cfBin)

# Copy binary to paths.RHESSYS_BIN
cfSrc = os.path.join(buildPath, cfBin)
cfDest = os.path.join(paths.RHESSYS_BIN, cfBin)
shutil.copyfile(cfSrc, cfDest)
os.chmod(cfDest, permissions)
# Write metadata
RHESSysMetadata.writeRHESSysEntry(context, 'cf_bin', paths.relpath(cfDest) )

# Build lairead
buildPath = os.path.join(paths.RHESSYS_SRC, 'util', 'GRASS', 'lairead')
makeCommand = makePath
process = Popen(makeCommand, shell=True, cwd=buildPath)
returnCode = process.wait()
if returnCode != 0:
    sys.exit("Failed to build lairead, return code: %s" % (returnCode,) )
# Copy binary to paths.RHESSYS_BIN
src = os.path.join(buildPath, 'lairead')
dest = os.path.join(paths.RHESSYS_BIN, 'lairead')
shutil.copyfile(src, dest)
os.chmod(dest, permissions)
# Write metadata
RHESSysMetadata.writeRHESSysEntry(context, 'lairead_bin', paths.relpath(dest) )

# Build RHESSys
buildPath = os.path.join(paths.RHESSYS_SRC, 'rhessys')
makeCommand = makePath
process = Popen(makeCommand, shell=True, cwd=buildPath)
returnCode = process.wait()
if returnCode != 0:
    sys.exit("Failed to build rhessys, return code: %s" % (returnCode,) )

# Find the RHESSys binary
rhessysBinRegex = re.compile('^rhessys.+$')
contents = os.listdir(buildPath)
rhessysBin = None
for entry in contents:
    m = rhessysBinRegex.match(entry)
    if m:
        rhessysBin = entry
        break
assert(rhessysBin)

# Copy binary to paths.RHESSYS_BIN
rhessysSrc = os.path.join(buildPath, rhessysBin)
rhessysDest = os.path.join(paths.RHESSYS_BIN, rhessysBin)
shutil.copyfile(rhessysSrc, rhessysDest)
os.chmod(rhessysDest, permissions)
# Write metadata
RHESSysMetadata.writeRHESSysEntry(context, 'rhessys_bin', paths.relpath(rhessysDest) )

# Write metadata
RHESSysMetadata.writeRHESSysEntry(context, 'exec_dir', paths.relpath(paths.RHESSYS_BIN) )

sys.stdout.write('\n\nFinished importing RHESSys source\n')

# Write processing history
RHESSysMetadata.appendProcessingHistoryItem(context, cmdline)


from setuptools import setup

def readme():
    with open('README.md') as f:
        return f.read()

setup(name='rhessysworkflows',
      version='1.34',
      description='Libraries and command-line scripts for performing RHESSys data preparation workflows.',
      long_description=readme(),
      classifiers=[
        'Development Status :: 5 - Production/Stable',
        'Environment :: Console',
        'Intended Audience :: Science/Research',
        'License :: OSI Approved :: BSD License',
        'Natural Language :: English',
        'Operating System :: Unix',
        'Topic :: Scientific/Engineering :: GIS'        
      ],
      url='https://github.com/selimnairb/RHESSysWorkflows',
      author='Brian Miles',
      author_email='brian_miles@unc.edu',
      license='BSD',
      packages=['rhessysworkflows',
                'rhessysworkflows.command',
                'rhessysworkflows.tests'
                ],
      install_requires=[
        'ecohydrolib>=1.27',
        'numpy>=1.7',
        'matplotlib>=1.1',
        'pandas',
        'scipy',
        'patsy',
        'statsmodels',
        'requests'
      ],
      scripts=['bin/CreateFlowtable.py',
               'bin/CreateFlowtableMultiple.py',
               'bin/CreateGRASSLocationFromDEM.py',
               'bin/CreateWorldfile.py',
               'bin/CreateWorldfileMultiple.py',
               'bin/DelineateWatershed.py',
               'bin/GenerateBaseStationMap.py',
               'bin/GenerateCustomSoilDefinitions.py',
               'bin/GenerateLandcoverMaps.py',
               'bin/GeneratePatchMap.py',
               'bin/GenerateSoilTextureMap.py',
               'bin/GenerateWorldTemplate.py',
               'bin/GIConverter.py',
               'bin/ImportClimateData.py',
               'bin/ImportRasterMapIntoGRASS.py',
               'bin/ImportRHESSysSource.py',
               'bin/PatchToCumulativeMap.py',
               'bin/PatchToCumulativeValues.py',
               'bin/PatchToMap.py',
               'bin/PatchToMovie.py',
               'bin/PatchZonalStats.py',
               'bin/PatchZonalStatsNormalize.py',
               'bin/RegisterCustomSoilReclassRules.py',
               'bin/RegisterLandcoverReclassRules.py',
               'bin/RHESSysPlot.py',
               'bin/RHESSysPlotMassbalance.py',
               'bin/RunLAIRead.py',
               'bin/RunLAIReadMultiple.py',
               'bin/RunModel.py'
      ],
      data_files=[('rhessysworkflows/etc/NLCD2006', ['etc/NLCD2006/impervious.rule',
                           'etc/NLCD2006/lai-recode.rule',
                           'etc/NLCD2006/landuse.rule',
                           'etc/NLCD2006/road.rule',
                           'etc/NLCD2006/stratum.rule'] ),
                  ('rhessysworkflows/etc/NLCD2011', ['etc/NLCD2011/impervious.rule',
                           'etc/NLCD2011/lai-recode.rule',
                           'etc/NLCD2011/landuse.rule',
                           'etc/NLCD2011/road.rule',
                           'etc/NLCD2011/stratum.rule'] ),
                  ('rhessysworkflows/etc/r.soils.texture', ['etc/r.soils.texture/FAO.dat',
                                                           'etc/r.soils.texture/isss.dat',
                                                           'etc/r.soils.texture/USDA.dat'] )
                  ],
                 
      zip_safe=False)



#!/usr/bin/env python
"""@package CreateWorldfile

@brief Generate world template to be used to create an initial worldfile 
for a RHESSys model.

This software is provided free of charge under the New BSD License. Please see
the following license information:

Copyright (c) 2013, University of North Carolina at Chapel Hill
All rights reserved.

Redistribution and use in source and binary forms, with or without
modification, are permitted provided that the following conditions are met:
    * Redistributions of source code must retain the above copyright
      notice, this list of conditions and the following disclaimer.
    * Redistributions in binary form must reproduce the above copyright
      notice, this list of conditions and the following disclaimer in the
      documentation and/or other materials provided with the distribution.
    * Neither the name of the University of North Carolina at Chapel Hill nor the
      names of its contributors may be used to endorse or promote products
      derived from this software without specific prior written permission.

THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND
ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
DISCLAIMED. IN NO EVENT SHALL THE UNIVERSITY OF NORTH CAROLINA AT CHAPEL HILL
BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR 
CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE
GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT 
LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT
OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.


@author Brian Miles <brian_miles@unc.edu>


Pre conditions
--------------
1. The following metadata entry(ies) must be present in the study area section of the metadata associated with the project directory: 
   bbox_wgs84

2. The following metadata entry(ies) must be present in the RHESSys section of the metadata associated with the project directory:
   stratum_defs
   landuse_defs
   soil_defs
   paramdb
   paramdb_dir
   rhessys_dir
   climate_stations
   template_template

3. The following metadata entry(ies) must be present in the GRASS section of the metadata associated with the project directory:
   basin_rast
   dem_rast
   hillslope_rast
   slope_rast
   aspect_rast
   zero_rast
   east_horizon_rast
   west_horizon_rast
   patch_rast
   soil_rast
   landuse_rast
   impervious_rast
   wetness_index_rast 
   stratum_rast
   xmap_rast
   ymap_rast
   
4. The following metadata entry(ies) will be used if present in the GRASS section of the metadata associated with the project directory:
   isohyet_rast
   basestation_rast
   
Post conditions
---------------
1. Template will be created in the RHESSys folder of the project directory.

2. Will write the following entry(ies) to the RHESSys section of metadata associated with the project directory:
   template

Usage:
@code
GenerateWorldTemplate.py -p /path/to/project_dir -c climate_station_name1 ... climate_station_nameN
@endcode

@note EcoHydroWorkflowLib configuration file must be specified by environmental variable 'ECOHYDROWORKFLOW_CFG',
or -i option must be specified. 
"""
import os, sys
import importlib
import string
import re
import argparse
import textwrap

from oset import oset

from ecohydrolib.grasslib import *
from ecohydrolib.spatialdata.utils import bboxFromString
from ecohydrolib.spatialdata.utils import calculateBoundingBoxCenter

from rhessysworkflows.context import Context
from rhessysworkflows.metadata import RHESSysMetadata
from rhessysworkflows.rhessys import RHESSysPaths
from rhessysworkflows.rhessys import readParameterFile

# Handle command line options
parser = argparse.ArgumentParser(description='Create RHESSys world template used to create initial world file')
parser.add_argument('-i', '--configfile', dest='configfile', required=False,
                    help='The configuration file. Must define section "GRASS" and option "GISBASE"')
parser.add_argument('-p', '--projectDir', dest='projectDir', required=True,
                    help='The directory to which metadata, intermediate, and final files should be saved')
parser.add_argument('-c', '--climateStation', dest='climateStation', required=False,
                     help='The climate station to associate with the worldfile.  Must be one of the climate stations specified in the "climate_stations" key in the "rhessys" section of the metadata')
parser.add_argument('--aspectMinSlopeOne', dest='aspectMinSlopeOne', action='store_true', required=False,
                    help='Use slope map with a minimum value of 1.0 to be used for calculating spherical average aspect.  Needed for areas of low slope due to limitations of RHESSys grass2world, which truncates slopes <1 to 0.0, which causes spherical average of aspect to equal NaN.')
parser.add_argument('-v', '--verbose', dest='verbose', action='store_true',
                    help='Print detailed information about what the program is doing')
args = parser.parse_args()
cmdline = RHESSysMetadata.getCommandLine()

configFile = None
if args.configfile:
    configFile = args.configfile

context = Context(args.projectDir, configFile) 

# Check for necessary information in metadata
studyArea = RHESSysMetadata.readStudyAreaEntries(context)
grassMetadata = RHESSysMetadata.readGRASSEntries(context)
if not 'dem_rast' in grassMetadata:
    sys.exit("Metadata in project directory %s does not contain a DEM raster in a GRASS mapset" % (context.projectDir,)) 
if not 'soil_rast' in grassMetadata:
    sys.exit("Metadata in project directory %s does not contain a soil raster in a GRASS mapset" % (context.projectDir,))
if not 'patch_rast' in grassMetadata:
    sys.exit("Metadata in project directory %s does not contain a patch raster in a GRASS mapset" % (context.projectDir,))

metadata = RHESSysMetadata.readRHESSysEntries(context)
if not 'grass_dbase' in metadata:
    sys.exit("Metadata in project directory %s does not contain a GRASS Dbase" % (context.projectDir,)) 
if not 'grass_location' in metadata:
    sys.exit("Metadata in project directory %s does not contain a GRASS location" % (context.projectDir,)) 
if not 'grass_mapset' in metadata:
    sys.exit("Metadata in project directory %s does not contain a GRASS mapset" % (context.projectDir,))
if not 'stratum_defs' in metadata:
    sys.exit("Metadata in project directory %s does not contain stratum definitions" % (context.projectDir,)) 
if not 'landuse_defs' in metadata:
    sys.exit("Metadata in project directory %s does not contain land use definitions" % (context.projectDir,)) 
if not 'soil_defs' in metadata:
    sys.exit("Metadata in project directory %s does not contain soil definitions" % (context.projectDir,)) 
if not 'rhessys_dir' in metadata:
    sys.exit("Metadata in project directory %s does not contain a RHESSys directory" % (context.projectDir,))
if not 'climate_stations' in metadata:
    sys.exit("Metadata in project directory %s does not contain a list of climate stations" % (context.projectDir,))
if not 'template_template' in metadata:
    sys.exit("Metadata in project directory %s does not contain a template template" % (context.projectDir,))
if not 'paramdb_dir' in metadata:
    sys.exit("Metadata in project directory %s does not contain a ParamDB directory" % (context.projectDir,))
if not 'paramdb' in metadata:
    sys.exit("Metadata in project directory %s does not contain a ParamDB" % (context.projectDir,))

if not args.climateStation and not 'basestations_rast' in grassMetadata:
    sys.exit("You must specify the climate station command line argument of generate a climate base station map")

climateStations = metadata['climate_stations'].split(RHESSysMetadata.VALUE_DELIM)
if args.climateStation and ( not args.climateStation in climateStations ):
    sys.exit("The chosen climate station '%s' was not found in the climate station list in metadata (%s)" %
             (str(args.climateStation), str(metadata['climate_stations']) ) )

rhessysDir = metadata['rhessys_dir']
paths = RHESSysPaths(args.projectDir, rhessysDir)

# Import ParamDB from project directory
paramDbPath = os.path.join(context.projectDir, metadata['paramdb'])
if not os.access(paramDbPath, os.R_OK):
    sys.exit("Unable to read RHESSys parameters database %s" % (paramDbPath,) )
sys.path.append( os.path.join(context.projectDir, metadata['paramdb_dir']) )
params = importlib.import_module('rhessys.params')
paramConst = importlib.import_module('rhessys.constants')
paramDB = params.paramDB(filename=paramDbPath)

# Set up GRASS environment
modulePath = context.config.get('GRASS', 'MODULE_PATH')
grassDbase = os.path.join(context.projectDir, metadata['grass_dbase'])
grassConfig = GRASSConfig(context, grassDbase, metadata['grass_location'], metadata['grass_mapset'])
grassLib = GRASSLib(grassConfig=grassConfig)

# Make sure mask and region are properly set
demRast = grassMetadata['dem_rast']
result = grassLib.script.run_command('g.region', rast=demRast)
if result != 0:
    sys.exit("g.region failed to set region to DEM, returning %s" % (result,))

# Make integer representation of DEM
dem1000Rast = "%s_1000" % (demRast,)
result = grassLib.script.write_command('r.mapcalc',
                                       stdin="%s=%s * 1000" % (dem1000Rast, demRast) )
if result != 0:
    sys.exit("r.mapcalc failed to integer DEM, returning %s" % (result,))

basinRast = grassMetadata['basin_rast']
result = grassLib.script.run_command('r.mask', flags='o', input=basinRast, maskcats='1')
if result != 0:
    sys.exit("r.mask failed to set mask to basin, returning %s" % (result,))

bbox = bboxFromString(studyArea['bbox_wgs84'])
(longitude, latitude) = calculateBoundingBoxCenter(bbox)

## 1. Get default files for basin, hillslope, and zone from default database
sys.stdout.write('Getting parameter definition files for basin, hillslope, and zone...')
sys.stdout.flush()
paramsFound = paramDB.search(paramConst.SEARCH_TYPE_CONSTRAINED, None, 'basin', None, None, None, None, None, None, None, None,
                             limitToBaseClasses=False, defaultIdOverride=str(1))
assert(paramsFound)
paramDB.writeParamFileForClass(paths.RHESSYS_DEF)
paramsFound = paramDB.search(paramConst.SEARCH_TYPE_CONSTRAINED, None, 'hillslope', None, None, None, None, None, None, None, None,
                             limitToBaseClasses=False, defaultIdOverride=str(1))
assert(paramsFound)
paramDB.writeParamFileForClass(paths.RHESSYS_DEF)
paramsFound = paramDB.search(paramConst.SEARCH_TYPE_CONSTRAINED, None, 'zone', None, None, None, None, None, None, None, None,
                             limitToBaseClasses=False, defaultIdOverride=str(1))
assert(paramsFound)
paramDB.writeParamFileForClass(paths.RHESSYS_DEF)
sys.stdout.write('done\n')

## 2. Determine the number of definition files of each type and save their names for inclusion in the world file
sys.stdout.write("\nFinding definition files in %s..." % (paths.RHESSYS_DEF,) )
sys.stdout.flush()
defFiles = {}
contents = os.listdir(paths.RHESSYS_DEF)
for type in paramConst.VALID_TYPES:
    if args.verbose:
        print("Definition files found for type: %s" % (type, ) )
    typeRe = re.compile("^%s_.+\.def$" % (type,) )
    if defFiles.has_key(type):
        defs = defFiles[type]
    else:
        defs = []
        defFiles[type] = defs
    for entry in contents:
        m = typeRe.match(entry)
        if m:
            if args.verbose:
                print("\t%s" % entry)
            defFiles[type].append(entry)

# Make sure there are soil, landuse and stratum defaults
if not defFiles.has_key('soil') or len(defFiles['soil']) < 1:
    sys.exit("No soil definition files found")
if not defFiles.has_key('landuse') or len(defFiles['landuse']) < 1:
    sys.exit("No land use definition files found")
if not defFiles.has_key('stratum') or len(defFiles['stratum']) < 1:
    sys.exit("No stratum definition files found")

sys.stdout.write('done\n')

## 3. Open worldfile template template and substitute values
templateTemplatePath = os.path.join(context.projectDir, metadata['template_template'])
sys.stdout.write("\nGenerating template from template template %s..." % (templateTemplatePath,) )
sys.stdout.flush()

templateTemplateFile = open(templateTemplatePath)
templateTemplate = string.Template(templateTemplateFile.read())
templateTemplateFile.close()
# Our dictionary of substitutions
subs = {}

# First, build substitution dictionary for default files
defs = None
for key in defFiles.keys():
    defs = defFiles[key]
    numDefin = len(defs)
    if numDefin >= 1:
        numDefinKey = "num_%s_defs" % (key,)    
        if args.verbose:
            print("%s: %d" % (numDefinKey, numDefin) )
        subs[numDefinKey] = str(numDefin)
    
        defStr = os.path.join(paths._DEF, defs[0])
        for defin in defs[1:]:
            defStr += os.linesep + os.path.join(paths._DEF, defin)
        defStrKey = "%s_defs" % (key,)
        if args.verbose:
            print("%s: %s" % (defStrKey, defStr) )
        subs[defStrKey] = defStr

# Second, climate stations
if args.climateStation:
    # Use the base station specified on command line
    if args.verbose:
        print("Using single climate station: %s" % (args.climateStation,) )
    climParamFilename = "%s.base" % (args.climateStation,)
    baseFile = os.path.join(paths._CLIM, climParamFilename)
    subs['climate_stations'] = baseFile
    subs['num_climate_stations'] = 1

    climParams = readParameterFile( os.path.join(paths.RHESSYS_CLIM, climParamFilename) )
    climateStationIDStr = "base_station_ID\tdvalue %s" % (climParams['base_station_id'],)
    subs['zone_base_station_ids'] = climateStationIDStr
    subs['zone_num_base_stations'] = 1
else:
    # Use base station raster map
    if args.verbose:
        print("Reading climate stations from raster: %s" % (grassMetadata['basestations_rast'],) )
    # Get list of base station IDs from raster
    rasterIds = oset()
    pipe = grassLib.script.pipe_command('r.stats', flags='licn', input=grassMetadata['basestations_rast'])
    for line in pipe.stdout:
        values = line.strip().split()
        if values[1] != 'NULL':
            rasterIds.add( int(values[0]) )
    # Get base station IDs from base station files
    baseIds = oset()
    baseFiles = {}
    for station in metadata['climate_stations'].split(','):
        climParamFilename = "%s.base" % (station,)
        baseFile = os.path.join( paths._CLIM, climParamFilename )
        climParams = readParameterFile( os.path.join(paths.RHESSYS_CLIM, climParamFilename) )
        id = int(climParams['base_station_id'])
        baseIds.add(id)
        baseFiles[id] = baseFile
    
    includedBaseIds = rasterIds & baseIds
    
    if len(includedBaseIds) == 0:
        sys.exit( textwrap.fill( "Some climate base station raster values do not have corresponding base station files. Raster values: %s, base station IDs: %s" % \
                                 (str(rasterIds), str(baseIds) ) ) )
    
    climateStationsStr = baseFiles[includedBaseIds[0]]
    for id in includedBaseIds[1:]:
        climateStationsStr += os.linesep + baseFiles[id]
    subs['climate_stations'] = climateStationsStr
    subs['num_climate_stations'] = len(includedBaseIds)
    
    climateStationIDStr = "base_station_ID\tmode %s" % (grassMetadata['basestations_rast'],)
    subs['zone_base_station_ids'] = climateStationIDStr
    subs['zone_num_base_stations'] = 1

# B. Everything else
subs['world_rast'] = grassMetadata['basin_rast']
subs['basin_rast'] = grassMetadata['basin_rast']
subs['dem_rast'] = dem1000Rast
subs['latitude_float'] = str(latitude)
subs['hillslope_rast'] = grassMetadata['hillslope_rast']
subs['zone_rast'] = grassMetadata['zone_rast']
subs['slope_rast'] = grassMetadata['slope_rast']
subs['aspect_rast'] = grassMetadata['aspect_rast']
if args.aspectMinSlopeOne:
    # Use minimum slope of 1 for calculating spherical average of aspect
    result = grassLib.script.write_command('r.mapcalc', stdin='slopegte1=if(slope<1,1,slope)')
    if result != 0:
        sys.exit("r.mapcalc failed to create slope >= 1 map, returning %s" % (result,))
    subs['aspect_slope_rast'] = 'slopegte1'
else:
    subs['aspect_slope_rast'] = grassMetadata['slope_rast']

if 'isohyet_rast' in grassMetadata:
    isohyet = grassMetadata['isohyet_rast']
    # Scale isohyet by 100 to avoid integer truncation by grass2world
    isohyet100 = isohyet + '100'
    result = grassLib.script.write_command('r.mapcalc', 
                                           stdin="{isohyet100} = {isohyet} * 100".format(isohyet100=isohyet100,
                                                                                         isohyet=isohyet))
    if result != 0:
        sys.exit("r.mapcalc failed to create %s map, returning %s" % (isohyet100, result))
    subs['isohyet_rast'] = isohyet100
else:
    result = grassLib.script.write_command('r.mapcalc', stdin='onehundred=100')
    if result != 0:
        sys.exit("r.mapcalc failed to create onehundred map, returning %s" % (result,))
    subs['isohyet_rast'] = 'onehundred'
subs['east_horizon_rast'] = grassMetadata['east_horizon_rast']
subs['west_horizon_rast'] = grassMetadata['west_horizon_rast']
subs['patch_rast'] = grassMetadata['patch_rast']
subs['soil_rast'] = grassMetadata['soil_rast']
subs['landuse_rast'] = grassMetadata['landuse_rast']
subs['impervious_rast'] = grassMetadata['impervious_rast']
subs['wetness_index_rast'] = grassMetadata['wetness_index_rast'] 
subs['stratum_rast'] = grassMetadata['stratum_rast']
subs['xmap_rast'] = grassMetadata['xmap_rast']
subs['ymap_rast'] = grassMetadata['ymap_rast']

# Third, substitute into the template template, producing the template, which we write to disk for g2w
templateStr = ''
try:
    templateStr = templateTemplate.substitute(subs)
    if args.verbose:
        print("Template:")
        print(templateStr)
except KeyError as e:
    sys.exit("ERROR creating worldfile template: template variable %s was not specified" % (str(e),) )
except ValueError:
    sys.exit("A '$' character was found in the template template, which is illegal")

# Write the template to a file stored in paths.RHESSYS_TEMPLATES, using the filename
# of the template template as a basis
templateFilename = os.path.splitext( os.path.split( metadata['template_template'] )[1] )[0]
templateFilepath = os.path.join(paths.RHESSYS_TEMPLATES, templateFilename)
f = open(templateFilepath, 'w')
f.write(templateStr)
f.close()
RHESSysMetadata.writeRHESSysEntry(context, 'template', paths.relpath(templateFilepath) )

sys.stdout.write('done\n')

# Write processing history
RHESSysMetadata.appendProcessingHistoryItem(context, cmdline)


#!/usr/bin/env python
"""@package RunModel

@brief Run RHESSys

This software is provided free of charge under the New BSD License. Please see
the following license information:

Copyright (c) 2013, University of North Carolina at Chapel Hill
All rights reserved.

Redistribution and use in source and binary forms, with or without
modification, are permitted provided that the following conditions are met:
    * Redistributions of source code must retain the above copyright
      notice, this list of conditions and the following disclaimer.
    * Redistributions in binary form must reproduce the above copyright
      notice, this list of conditions and the following disclaimer in the
      documentation and/or other materials provided with the distribution.
    * Neither the name of the University of North Carolina at Chapel Hill nor the
      names of its contributors may be used to endorse or promote products
      derived from this software without specific prior written permission.

THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND
ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
DISCLAIMED. IN NO EVENT SHALL THE UNIVERSITY OF NORTH CAROLINA AT CHAPEL HILL
BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR 
CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE
GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT 
LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT
OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.


@author Brian Miles <brian_miles@unc.edu>
  
  
Pre conditions
--------------
1. The following metadata entry(ies) must be present in the RHESSys section of the metadata associated with the project directory:
   rhessys_dir
   rhessys_bin

Post conditions
---------------
1. Will write an entry to the model run section of the project metadata

2. Will write an entry to the history section of the project metadata when the model run sucessfully completes

Usage:
@code
RunModel.py -p /path/to/project_dir -d "Scenario 1" (--basin | --hillslope | --zone | --patch | --canopy) -pre OUTPUT_PREFIX -st YYYY M D H -ed YYYY M D H -w WORLDFILE -t TECFILE -r FLOWTABLE [SURFACE_FLOWTABLE] [-- RHESSYS_ARG_1 ... RHESSYS_ARG_N]
@endcode
"""
import os
import sys
import errno
import argparse
import textwrap
import datetime
import subprocess

import ecohydrolib.util

from rhessysworkflows.context import Context
from rhessysworkflows.metadata import RHESSysMetadata
from rhessysworkflows.metadata import ModelRun
from rhessysworkflows.rhessys import RHESSysPaths
from rhessysworkflows.rhessys import generateCommandString

# Handle command line options
parser = argparse.ArgumentParser(description='Run RHESSys, recording information about the run in metadata')
parser.add_argument('-i', '--configfile', dest='configfile', required=False,
                    help='The configuration file')
parser.add_argument('-p', '--projectDir', dest='projectDir', required=True,
                    help='The directory to which metadata, intermediate, and final files should be saved')
parser.add_argument('-v', '--verbose', dest='verbose', action='store_true',
                    help='Print detailed information about what the program is doing')
parser.add_argument('-d', '--description', dest='description', required=True,
                    help='Description of the model run')
parser.add_argument('-pre', dest='outputPrefix', required=True,
                    help='Filename prefix to use for output files, relative to output directory in the RHESSys directory of the project')
parser.add_argument('-st', dest='startDate', required=True, nargs=4, type=int,
                    help='Start date and time of the model run, of the form "YYYY M D H"')
parser.add_argument('-ed', dest='endDate', required=True, nargs=4, type=int,
                    help='Date date and time of the model run, of the form "YYYY M D H"')
parser.add_argument('-w', dest='worldfile', required=True,
                    help='Filename of the worldfile to use for the model run, specified relative to the worldfiles directory in the RHESSys directory of the project')
parser.add_argument('-t', dest='tecfile', required=True,
                    help='Filename of the tecfile to use for the model run, specified relative to the tec directory in the RHESSys directory of the project')
parser.add_argument('-r', dest='flowtables', required=False, nargs='*',
                    help='Filename(s) of the flow table(s) to use for the model run, specified relative to the flowtable directory in the RHESSys directory of the project. ' +
                         'If one flow table is supplied, it will be used for subsurface and surface routing.  ' +
                         'If two flow tables are supplied the first will be use for subsurface routing, the second for surface.')
outputType = parser.add_mutually_exclusive_group(required=True)
outputType.add_argument('--basin', dest='outputType', action='store_const', const='-b',
                        help='Tell RHESSys to output at the basin spatial level')
outputType.add_argument('--hillslope', dest='outputType', action='store_const', const='-h',
                        help='Tell RHESSys to output at the hillslope spatial level')
outputType.add_argument('--zone', dest='outputType', action='store_const', const='-z',
                        help='Tell RHESSys to output at the zone spatial level')
outputType.add_argument('--patch', dest='outputType', action='store_const', const='-p',
                        help='Tell RHESSys to output at the patch spatial level')
outputType.add_argument('--canopy', dest='outputType', action='store_const', const='-c',
                        help='Tell RHESSys to output at the canopy stratum spatial level')
parser.add_argument('args', nargs=argparse.REMAINDER)
args = parser.parse_args()
cmdline = RHESSysMetadata.getCommandLine()

configFile = None
try:
    configFile = args.configfile
except AttributeError:
    pass

context = Context(args.projectDir, configFile) 

# Check for necessary information in metadata
metadata = RHESSysMetadata.readRHESSysEntries(context)
if not 'rhessys_dir' in metadata:
    sys.exit("Metadata in project directory %s does not contain a RHESSys directory" % (context.projectDir,))
if not 'rhessys_bin' in metadata:
    sys.exit("Metadata in project directory %s does not contain a RHESSys binary" % (context.projectDir,))

# Make sure specified files exist
rhessysDir = metadata['rhessys_dir']
paths = RHESSysPaths(args.projectDir, rhessysDir)

rhessysBinPathRel = metadata['rhessys_bin']
rhessysBinPath = os.path.join(context.projectDir, rhessysBinPathRel)
if not os.path.isfile(rhessysBinPath) or not os.access(rhessysBinPath, os.X_OK):
    sys.exit("Putative RHESSys executable '%s' is not an executable file" % (rhessysBinPath,) )

if args.outputPrefix.count(os.sep) > 1:
    sys.exit("Output prefix cannot contain a path separator ('%s')" % (args.outputPrefix,) )
if not os.access(paths.RHESSYS_OUT, os.W_OK):
    sys.exit("RHESSys output directory '%s' is not writable" % (paths.RHESSYS_OUT,) )
outputDir = os.path.join(paths.RHESSYS_OUT, args.outputPrefix)
# Append path separator and 'rhessys' so that RHESSys will write output into a subdirectory
outputPrefix = os.path.join(outputDir, RHESSysMetadata.MODEL_NAME)
if os.path.isdir(outputPrefix):
    sys.exit( textwrap.fill("\nOutput directory '%s' already exists, please choose another output prefix." % (outputPrefix,) ) )
outputPrefixRel = paths.relpath(outputPrefix)
outputDirRel = paths.relpath(outputDir)

if args.worldfile.count(os.sep) > 1:
    sys.exit("Worldfile cannot contain a path separator ('%s')" % (args.worldfile,) )
worldfile = os.path.join(paths.RHESSYS_WORLD, args.worldfile)
if not os.path.isfile(worldfile) or not os.access(worldfile, os.R_OK):
    sys.exit("Worldfile '%s' is not a readable file" % (worldfile,) )
worldfileRel = paths.relpath(worldfile)
    
if args.tecfile.count(os.sep) > 1:
    sys.exit("TEC file cannot contain a path separator ('%s')" % (args.tecfile,) )
tecfile = os.path.join(paths.RHESSYS_TEC, args.tecfile)
if not os.path.isfile(tecfile) or not os.access(tecfile, os.R_OK):
    sys.exit("TEC file '%s' is not a readable file" % (tecfile,) )
tecfileRel = paths.relpath(tecfile)

if args.flowtables:
    # We are running in routing mode
    cmd_proto = "{bin} -st {startDate} -ed {endDate} {outputType} -pre {outputPrefix} -t {tecfile} -w {worldfile} -r {flowtables} {remainders}"

    surfaceFlow = None
    subsurfaceFlow = args.flowtables[0]
    if subsurfaceFlow.count(os.sep) > 1:
        sys.exit("Flowtable cannot contain a path separator ('%s')" % (subsurfaceFlow,) )
    subsurfaceFlow = os.path.join(paths.RHESSYS_FLOW, subsurfaceFlow)
    if not os.path.isfile(subsurfaceFlow) or not os.access(subsurfaceFlow, os.R_OK):
        sys.exit("Flowtable '%s' is not a readable file" % (subsurfaceFlow,) )
    subsurfaceFlowRel = paths.relpath(subsurfaceFlow)
    flowtables = subsurfaceFlow
    flowtablesRel = subsurfaceFlowRel
        
    if len(args.flowtables) == 2:
        surfaceFlow = args.flowtables[1]
        if surfaceFlow.count(os.sep) > 1:
            sys.exit("Surface flowtable cannot contain a path separator ('%s')" % (surfaceFlow,) )
        surfaceFlow = os.path.join(paths.RHESSYS_FLOW, surfaceFlow)
        if not os.path.isfile(surfaceFlow) or not os.access(surfaceFlow, os.R_OK):
            sys.exit("Surface flowtable '%s' is not a readable file" % (surfaceFlow,) )
        surfaceFlowRel = paths.relpath(surfaceFlow)
        flowtables += ' ' + surfaceFlow
        flowtablesRel += ' ' + surfaceFlowRel
else:
    # We are running in topmodel mode
    cmd_proto = "{bin} -st {startDate} -ed {endDate} {outputType} -pre {outputPrefix} -t {tecfile} -w {worldfile} {remainders}"
    flowtables = flowtablesRel = None

remainders = ' '.join(args.args[1:])
startDate = ' '.join([str(d) for d in args.startDate])
endDate = ' '.join([str(d) for d in args.endDate])

# Build command string for running (i.e. with absolute paths)
cmd = cmd_proto.format(bin=rhessysBinPath, startDate=startDate, endDate=endDate,
                       outputType=args.outputType, outputPrefix=outputPrefix,
                       tecfile=tecfile, worldfile=worldfile, flowtables=flowtables, remainders=remainders)
# Build command string with relative paths for metadata
cmdRel = cmd_proto.format(bin=rhessysBinPathRel, startDate=startDate, endDate=endDate,
                       outputType=args.outputType, outputPrefix=outputPrefixRel,
                       tecfile=tecfileRel, worldfile=worldfileRel, flowtables=flowtablesRel, remainders=remainders)

# Run RHESSys
sys.stdout.write("Running RHESSys...")

cmdArgs = cmd.split()
process = subprocess.Popen(cmdArgs, cwd=paths.RHESSYS_DIR, stdout=subprocess.PIPE, stderr=subprocess.STDOUT)

# Create output directory so we can store RHESSys output there
os.makedirs(outputDir)
rhessysOutPath = os.path.join(outputDir, 'rhessys.out')
rhessysOut = open(rhessysOutPath, 'w')

# Get output
while True:
    line = process.stdout.readline()
    if not line:
        break
    rhessysOut.write(line)
    if args.verbose:
        sys.stdout.write(line)

if process.wait() != 0:
    sys.exit("\n\nRHESSys failed, returning %s" % (process.returncode,) )

rhessysOut.close() 
print('\n\nRHESSys model run successfully completed\n')

# Write metadata about run
run = ModelRun()
run.description = args.description
run.date = datetime.datetime.utcnow()
run.command = cmdRel
run.output = outputDirRel
run.writeToMetadata(context)

# Write processing history
RHESSysMetadata.appendProcessingHistoryItem(context, cmdline)

